package de.uni_leipzig.asv.unsupos;

import java.io.*;

/**
 * This is the application class of the unsupervised Part-of-Speech (POS) tagger
 * from Chris Biemann (Leipzig University).
 * 
 * @author  Andreas Klaus
 * @version 0.1
 */
public class UnsuPos {
	private static PrintStream sout;
	private static PrintStream eout;
	private static String fcfname = "./config/unsupos.conf";
	private static File fcorpus;
	private static boolean ignoreCase = false;
	
    /**
     * Reads and checks the command line arguments specified by the args argument.
     * <p>
     * -input corpus-file.txt (mandatory)
     * <p>
     * -config config-file.txt (optional, the default configuration file is
     *                          ./config/unsupos.conf)
     * <p>
     * Note: corpus-file.txt is the text file to be tagged (one sentence per line).
     * 
     * @param  args the command line arguments
     * @throws Exception
     */
    public static void scanParams(String[] args) throws Exception {
    	
    	/* Scan for help. */
    	boolean printHelp = (args.length == 0);
    	String  helpArg   = "-help";
    	for (int i = 0; i < args.length; i++) {
    		if ((args[i].compareToIgnoreCase("help")   == 0) || 
    			(args[i].compareToIgnoreCase("-help")  == 0) ||
    			(args[i].compareToIgnoreCase("--help") == 0) ||
    			(args[i].compareToIgnoreCase("-h")     == 0) ||
    			(args[i].compareToIgnoreCase("?")      == 0)) {
    			printHelp = true;
    			if (args[i].compareTo("-help") != 0) {
    				helpArg = helpArg + ", " + args[i]; }
    			break; } }
    	
    	if (printHelp) {
    		System.out.println("");
    		System.out.println("This program invokes the unsupervised Part-of-Speech tagger.");
    		System.out.println("");
    		System.out.println("ARGUMENTS");
    		System.out.println("  " + helpArg);
    		System.out.println("               prints this information");
    		System.out.println("  -input  CORPUSFILE");
    		System.out.println("               uses corpus file CORPUSFILE (one sentence per");
    		System.out.println("               line)");
    		System.out.println("  -config PARAMETERFILE");
    		System.out.println("               uses parameters from PARAMETERFILE");
    		System.out.println("  -ignoreCase");
    		System.out.println("               run the program case insensitive");
    		System.out.println("------------------------------------------------------------");
    		System.out.println("For further information on Java configuration see README.txt");
    		System.out.println("");
    		System.out.println("EXAMPLE Java options are");
    		System.out.println("  -Xmx712M     maximum size of the memory allocation pool");
    		System.out.println("  -cp \".;./lib/*\"");
    		System.out.println("               specify class path entries");
    		System.out.println("  -Dde.uni_leipzig.asv.medusa.config.ClassConfig=");
    		System.out.println("                      ./config/medusa_config.xml");
    		System.out.println("               specify Medusa configuration file");
    		System.out.println("------------------------------------------------------------");
    		throw new Exception(); }
    	
    	/* Scan all arguments. */
        for (int i = 0; i < args.length; i++) {
       	    
        	/* Corpus file. */
        	if (args[i].compareTo("-input") == 0) {
    			if ((i + 1) < args.length) {
    			  fcorpus = new File(args[i+1]); }
    			else {
    				throw new Exception("Argument error: no corpus file specified."); }
    			
    			/* Test if file exisits. */
    			if (!fcorpus.exists() || !fcorpus.isFile()) {
    				throw new Exception("File " + fcorpus.getPath() + " can not be found."); } }
        	
    		/* Test for configuration file name. */
    		if (args[i].compareTo("-config") == 0) {
    			if ((i + 1) < args.length) {
    			  fcfname = args[i+1]; }
    			else {
    				throw new Exception("Argument error: no config file specified."); } }
    		
    		/* Case sensitivity. */
    		if (args[i].compareTo("-ignoreCase") == 0) {
    			ignoreCase = true; }
    		else {
    			ignoreCase = false; } }
    	return; }

    /**
     * This is the entry point of the application.
     * 
	 * @param args the command line arguments
	 * @see   #scanParams(String[]) <code>scanParams(String[])</code>
	 */
	public static void main(String[] args) {
		
		/* Scan and check command line arguments. */
		try {
			scanParams(args); }
		catch (Exception e) {
			String msg = "";
			if (e.getMessage() != null) { msg = e.getMessage(); }
			System.err.println(msg);
			System.exit(0); }
		
		/* Create output PrintStream for standard output and standard error. */
		try {
			/* Get directory of corpus file. */
			String dir = new File(fcorpus.getAbsolutePath()).getParent(); 
			File logdir = new File(dir + "/log/");
			if ((!logdir.exists() && !logdir.mkdir()) || !logdir.isDirectory()) {
				 System.err.println("Cannot create log directory " + logdir.getPath());
				 System.exit(-1); }
			File fs = new File(logdir.getPath(), fcorpus.getName() + ".stdout");
			File fe = new File(logdir.getPath(), fcorpus.getName() + ".stderr");
			sout = new PrintStream(new FileOutputStream(fs));
			eout = new PrintStream(new FileOutputStream(fe));
			System.out.println("Write Unsupos standard output to " + fs.getPath());
			System.out.println("Write Unsupos standard error to "  + fe.getPath()); }
		catch (FileNotFoundException e) {
			System.err.println("Cannot open streams for standard output and standard error.");
			System.exit(-1); }
		
		/* Initialize and start the unsupervized PoS tagger. */
		try {
			UnsuPosTagger upost = new UnsuPosTagger(fcorpus.getPath(), ignoreCase, fcfname, sout, eout, null);
			upost.printConfigInfo();
			Thread p = new Thread(upost);
			p.start();
			
			BufferedWriter fstate = new BufferedWriter(new FileWriter("state.log"));
			while (p.isAlive()) {
				fstate.write(upost.getActualStepName() + " - " +
							 upost.getActualTaskName() + ": "  +
							 upost.getActualTaskProgress());
				fstate.newLine();
				fstate.flush();
				Thread.sleep(10000); }
			fstate.close();
			
			/* For getting the task state the following loop can be used.
			 *     while (p.isAlive()) {
			 *         ... = upost.getActualStepName();
			 *         ... = upost.getActualTaskName();
			 *         ... = upost.getActualTaskProgress();
			 *     }
			 */
			
			p.join(); }
		catch (Exception e) {
			System.err.println("Exception in the unsupervised Part-of-Speech tagger");
			System.err.println("====================================================");
			System.err.println(e.getMessage());
			System.err.println("----------------------------------------------------"); }
		System.out.println("Bye ...");
		return; }
	
}
