package de.uni_leipzig.asv.unsupos;

import java.io.*;
import java.util.*;
import java.lang.Math;
import de.uni_leipzig.asv.medusa.controlflow.AbstractControlFlowImpl;
import de.uni_leipzig.asv.medusa.config.ConfigurationException;
import de.uni_leipzig.asv.medusa.config.ConfigurationContainer;
import de.uni_leipzig.asv.medusa.hash.HashException;
import de.uni_leipzig.asv.medusa.input.ImprovedWordTokenizer;
import de.uni_leipzig.asv.medusa.input.WordNumbers;
import de.uni_leipzig.asv.medusa.input.SentenceToIdx2;
import de.uni_leipzig.asv.medusa.transform.Transform;
import de.uni_leipzig.asv.medusa.transform.WordNumberTransformerImpl;
import de.uni_leipzig.asv.filesort.FileSort;
import de.uni_leipzig.asv.toolbox.ChineseWhispers.algorithm.ChineseWhispers;
import de.uni_leipzig.asv.unsupos.similarity.SimOptions;
import de.uni_leipzig.asv.unsupos.similarity.Similarity;
import de.uni_leipzig.asv.coocc.BinFileConstructor;
import de.uni_leipzig.asv.coocc.BinFileMultCol;

/**
 * This class implements the unsupervised the Part-of-Speech (POS) tagger from Chris Biemann
 * (University of Leipzig).<p>
 * 
 * <strong>Reference paper:</strong> Biemann C. (2006): Unsupervised Part-of-Speech Tagging Employing
 *                                   Efficient Graph Clustering. Proceedings of the COLING/ACL-06 Student
 *                                   Research Workshop 2006, Sydney, Australia.
 * 
 *  @author  Andreas Klaus
 *  @version 1.0
 */
public class UnsuPosTagger extends AbstractControlFlowImpl implements Runnable {

	/**
	 * The <code>MyCosRun</code> class is used to calculate the cosinus of the angle between two
	 * context vectors for the cosinus similarity graph (partitioning 1).
	 */
	private class MyCosRun implements Runnable {
		
		/**
		 * Constructs a new runnable <code>MyCosRun</code>.
		 *  
		 * @param offset the vector index to start with.
		 * @param inc    the increment (i.e. the number of threads).
		 * @param fout   the buffered writer for output.
		 */
		public MyCosRun(int offset, int inc, BufferedWriter fout) {
			this.offset       = offset;
			this.inc          = inc;
			this.fout         = fout;
			this.taskProgress = 0;
			return; }
		
		/**
		 * This method implements <code>java.lang.Runnable.run()</code> and starts the cosinus calculation
		 * for vectors specified by <code>offset</code> and <code>inc</code>. Output (graph edges) is
		 * written to <code>fout</code>.
		 */
		public void run() {
			double oneplus = 1.0 + 0.000000001;
			
			try {
				int update1 = Math.max(1, (int) (Nwtarg / (100 * inc)));
				int update2 = Math.max(1, (int) (Nwtarg / 50));
				int n = 0;
				taskProgress = 0;
				double vik, vjk;
				for (int i = offset; i <= Nwtarg; i += inc) {
					if ((n   % update1) == 0) {
						taskProgress = Math.min(100, taskProgress + 1); }
					if ((n++ % update2) == 0) {
						printInfo("."); }
					
					/* Calculate the scalar product between vector i and j. */
					for (int j = i+1; j <= Nwtarg; j++) {
						double sp = 0.0;
						for (int k = 0; k < 4 * Nwfeat; k++) {
							vik = (double) V[i][k];
							vjk = (double) V[j][k];
							sp += vik * vjk; }
					
						/* Cosinus of angle between context vector i and j.
						 * Note: Here, the norm of a vector is equal to zero, or greater than or
						 *       equal to one. 
						 */
						if ((vnorm[i] * vnorm[j]) > 0.5) {
							double cs = sp / (vnorm[i] * vnorm[j]);
							if (cs >= Tcos) {
								int w = (int) (10.0 / (oneplus - cs));
								fout.write(i + "\t" + j + "\t" + w + NL);
								fout.write(j + "\t" + i + "\t" + w + NL); } } } } }
			catch (IOException e) { }
			return;	}
		
		
		/**
		 * The first vector index to start with. Next vector index is <code>offset + inc</code>, and so on.
		 */
		private int offset;
		
		/**
		 * The increment (should correspond to the number of <code>MyCosRun</code>-threads started).
		 */
		private int inc;
		
		/**
		 * The writer for buffered output of graph edges.
		 */
		private BufferedWriter fout;
		
		/**
		 * The progress of the cosinus calculation (between 0 and 100).
		 */
	    public int taskProgress; }
	
	/**
	 * This class calculates the cooccurrences in the corpus.
	 */
	private class MyCooccRun extends AbstractControlFlowImpl {
		
		/**
		 * @throws Exception
		 */
		public MyCooccRun() throws Exception {
			taskProgress = 0;
			try {
				init(); }
			catch (ConfigurationException e) {
				throw new Exception(e); }
			return; }

		/**
		 * This method calculates the cooccurrences.
		 * 
		 * @throws Exception
		 */
		public void start() throws Exception {
			taskProgress = 2;
			try {
				/*printlnInfo("Medusa: Transform word numbers.");
				printInfoAlienOutputBegin("Medusa output");
				setGlobalProperty("MEMORY_ALLOCATOR_IMPL",
	                              "de.uni_leipzig.asv.medusa.config.DefaultMemoryAllocatorImpl");
				Transform trans = new WordNumberTransformerImpl();
				trans.init();
				trans.transform();
				printInfoAlienOutputEnd();*/
				taskProgress += 25;
				
				printlnInfo("Medusa: Calculate sentence index file.");
				printInfoAlienOutputBegin("Medusa output");
				SentenceToIdx2 objIDX = new SentenceToIdx2();
				objIDX.init();
				objIDX.startReading();
				objIDX.finish();
				objIDX = null;
				printInfoAlienOutputEnd();
				taskProgress += 45;
				
				printlnInfo("Medusa: Count words/frequencies.");
				printInfoAlienOutputBegin("Medusa output");
				setGlobalProperty("PARSER_FILTER_IMPL",
						 "de.uni_leipzig.asv.medusa.loader.filter.IDXDistinctWordFrequencyCounterFilterImpl");
				loadParser();
				objLoader.parse();
				objLoader.store();
				objLoader = null;
				printInfoAlienOutputEnd();
				taskProgress +=  5;
				
				printlnInfo("Medusa: Allocate hash memory.");
				printInfoAlienOutputBegin("Medusa output");
				loadMemoryAllocator();
				objMemAllo.allocateHash();
				printInfoAlienOutputEnd();
				taskProgress +=  2;
				
				printlnInfo("Medusa: Run filter.");
				printInfoAlienOutputBegin("Medusa output");
				runFilterComponent();
				printInfoAlienOutputEnd();
				taskProgress +=  1;
				
				printlnInfo("Medusa: Parse corpus file.");
				printInfoAlienOutputBegin("Medusa output");
				setGlobalProperty("PARSER_FILTER_IMPL",
						 "de.uni_leipzig.asv.medusa.loader.filter.IDXNeighbourhoodFilterImpl");
				loadParser();
				objLoader.parse();
				objLoader.store();
				objLoader = null;
				printInfoAlienOutputEnd();
				taskProgress += 10;
				
				printlnInfo("Medusa: Export results.");
				printInfoAlienOutputBegin("Medusa output");
				loadExporter();
				objExporter.export();
				objExporter = null;
				printInfoAlienOutputEnd();
				taskProgress += 10; }
			catch (ConfigurationException e) {
				throw new Exception(e); }
			catch (HashException e) {
				throw new Exception(e); }
			catch (Exception e) {
				throw new Exception(e); }
			return;	} }
	
	/**
	 * Performs full similarity: how many 2-paths are existing between two nodes in a graph.
	 * 
	 * @author Chris Biemann
	 */
	@SuppressWarnings("unused")
	private class MySimBiem implements Runnable {
		public MySimBiem(String filename12, String filename21, boolean graphInMemory, String fedgeName) {
			this.edges12   = BinFileConstructor.getBinFileMultCol(filename12, 3, graphInMemory);
	        this.edges21   = BinFileConstructor.getBinFileMultCol(filename21, 3, graphInMemory);
	        this.fedgeName = fedgeName;
	        this.taskProgress = 0; }
		
		/**
		 * This method calculates a symmetric NB graph. 
		 */
		@SuppressWarnings("unchecked")
		public void run() {
			try {
				
				/* Create a temporary list with possibly non-unique edges. */
				File ftmp = File.createTempFile(fcorpus.getName() + ".graph2.nu-edges", null);
				tmpFileList.add(ftmp.getAbsolutePath());
				BufferedWriter bftmp = new BufferedWriter(new FileWriter(ftmp.getPath()));
				
				Hashtable secondorder;
		        int thresh    = Tnb;
		        int startfrom = Nbehead;
	
		        int min12=edges12.getFirstNonNullPosition().intValue();
		        int max12=edges12.getMaxIndexPosition().intValue();
		        int min21=edges21.getFirstNonNullPosition().intValue();
		        int max21=edges21.getMaxIndexPosition().intValue();
		        int max=0;
		        int min=0;
	
		        if (max12>max21) {max=max21;}else {max=max12;}
		        if (min12>min21) {min=min12;}else {min=min21;}
		        if (min<startfrom) {min=startfrom;}

	        	for(int i=min;i<max+1;i++) {
	        		
	        		/* Set task progress. */
	        		if (((i-min) % ((max+1-min) / 100)) == 0) {
	        			this.taskProgress = Math.min(100, this.taskProgress + 1); }
	        		if (((i-min) % ((max+1-min) / 50)) == 0) {
	        			printInfo("."); }
	        		
	        		secondorder = two_paths(new Integer(i), thresh, startfrom);
	        		for (Enumeration e=secondorder.keys();e.hasMoreElements();) {
	        			Integer kkey=(Integer)e.nextElement();
	        			int j   = kkey.intValue();
	        			int val = ((Integer)secondorder.get(kkey)).intValue();
	        			if (i!=j) { // no self-sim
	        				
	        				/* Write undirected nb edge. */
	        				bftmp.write(i + "\t" + j + "\t" + val + NL);
	        				bftmp.write(j + "\t" + i + "\t" + val + NL);
	        				/*System.out.println(i+"\t"+kkey+"\t"+val);*/ } } }
	        	bftmp.close();
	        	removeNonUniqueEdges(ftmp.getPath(), fedgeName); }
	        catch (IOException e) {
	        	printlnError("Error while writing graph 2.");
	        	stopHere("Error while writing graph 2."); }
	        printlnInfo("");
	        return; }
		
		private void removeNonUniqueEdges(String fnuedgesName, String fedgesName) {
			try {
				File ftmp = File.createTempFile(fcorpus.getName() + ".graph2.nu-edges.sort", null);
				tmpFileList.add(ftmp.getAbsolutePath());
				
				/* Sort edge list. */
				int[]  sOrder = {  0,   1,   2};
				char[] sType  = {'i', 'i', 'I'};
				FileSort s = new FileSort("\t", sOrder, sType);
				s.sort(fnuedgesName, ftmp.getPath());
				
				/* Remove non-unique edges from list. */
				BufferedWriter fout = new BufferedWriter(new FileWriter(fedgesName));
				BufferedReader fin  = new BufferedReader(new FileReader(ftmp));
				String line, e0 = "-1", e1 = "-1";
				while ((line = fin.readLine()) != null) {
					String[] edge = line.split("\t");
					if (edge.length != 3) {
						continue; }
					if ((edge[0].compareTo(e0) != 0) || (edge[1].compareTo(e1) != 0)) {
						e0 = edge[0];
						e1 = edge[1];
						fout.write(line + NL); } }
				fout.close();
				fin.close(); }
			catch (IOException e) {
				printlnError("Error while sorting edge list for graph 2.");
	        	stopHere("Error while sorting edge list for graph 2."); }
			return;	}
		
		@SuppressWarnings("unchecked")
		private Hashtable two_paths(Integer node, int thresh, int startfrom) {
	        Hashtable table12= new Hashtable();
	        Hashtable table21= new Hashtable();
	        Hashtable rettable=new Hashtable();

	        // edges12
	        // get 1st-order
	        List list=(List)edges12.getSingleColumn(node, Nnbmax, 2);
	        for (Iterator it = list.iterator(); it.hasNext();) {
	            List second = (List)edges21.getSingleColumn((Integer)it.next(), Nnbmax, 2);
	            for (Iterator sit = second.iterator(); sit.hasNext();) {
	                Integer target=(Integer)sit.next();
	                if (target.intValue()>=startfrom) {
	                    if (table12.containsKey(target)) {
	                        table12.put(target,new Integer(((Integer)table12.get(target)).intValue()+1)); }
	                    else {
	                        table12.put(target, new Integer(1)); } } } }
	        for (Enumeration e=table12.keys();e.hasMoreElements();) {
	            Integer kkey=(Integer)e.nextElement();
	            int val=((Integer)table12.get(kkey)).intValue();
	            if (val<thresh) { table12.remove(kkey);} }

	        // edges21, get 1st-order
	        list=(List)edges21.getSingleColumn(node, Nnbmax, 2);
	        for (Iterator it = list.iterator(); it.hasNext();) {
	            List second = (List)edges12.getSingleColumn((Integer)it.next(), Nnbmax, 2);
	            for (Iterator sit = second.iterator(); sit.hasNext();) {
	                Integer target=(Integer)sit.next();
	                if (target.intValue()>=startfrom) {
	                    if (table21.containsKey(target)) {
	                        table21.put(target,new Integer(((Integer)table21.get(target)).intValue()+1)); }
	                    else {
	                        table21.put(target, new Integer(1)); } } } }

	        // filter table 2 and join
	        for (Enumeration e=table21.keys();e.hasMoreElements();) {
	            Integer kkey=(Integer)e.nextElement();
	            int val=((Integer)table21.get(kkey)).intValue();
	            if (val>=thresh) {
	                if (table12.containsKey(kkey)) {
	                    rettable.put(kkey,new Integer(val+((Integer)table12.get(kkey)).intValue())); } } }
	        return rettable; }
		
		public int getTaskProgress() {
			return this.taskProgress; }
		
		private int taskProgress = 0;
		private BinFileMultCol edges12;
	    private BinFileMultCol edges21;
	    private String fedgeName; }
	
	/**
	 * This class calculates a similarity graph. It uses an adapted version of the
	 * <code>Similarity</code>-package from the ASV group (University of Leipzig).
	 * 
	 * @deprecated
	 */
	private class MySimRun implements Runnable {
		
		/**
		 * Creates a similarity calculator and prepares cue and feature data.
		 * 
		 * @param fsymname the file name of the symmetric neighbourhood graph.
		 * @param fnbsym   the file name of the neighbourhood graph.
		 * @param foname   the output file name for the similarity graph.
		 */
		public MySimRun(String fsymname, String fnbname, String foname) throws Exception {
			
			/* Configuration. */
			simopt = new SimOptions();
			simopt.setLoadCueToRam(false);
			simopt.setLoadFeaturesToRam(false);
			simopt.setOutputMainString(foname);
			
			/* Prepare cue data. */
			cueData = BinFileConstructor.getBinFileMultCol(fsymname, 3, simopt.getLoadCueToRam());
			
			/* Prepare feature data. */
			featureData = new BinFileMultCol[1];
			featureData[0] = BinFileConstructor.getBinFileMultCol(fnbname, 3, simopt.getLoadFeaturesToRam());
			
			/* The similarity calculator. */
			sim = new Similarity(simopt, cueData, featureData);
			return;	}
		
		/**
		 * This method starts the similarity calculation (implements <code>java.lang.Runnable.run()</code>).
		 */
		public void run() {
			sim.runAlgorithm();
			return;	}
		
		/**
		 * This method returns the actual progress of the similarity calculation.
		 * 
		 * @return the actual progress of the calculation (between 0 and 100).
		 */
		public int getTaskProgress() {
			return sim.taskProgress; }
		
		/**
		 * The cue data used for the similarity calculation.
		 */
		private BinFileMultCol cueData;
		
		/**
		 * The feature data used for the similarity calculation.
		 */
		private BinFileMultCol[] featureData;;
		
		/**
		 * <code>Similarity</code> configuration.
		 */
		private SimOptions simopt;
	
		/**
		 * The <code>Similarity</code> algorithm to run.
		 */
		private Similarity sim; }
	
	/**
	 * The constructor for the unsupervized Part-of-Speech tagger. It reads the configuration file and
	 * the corpus file to determine the number of sentences.<p>
	 *
	 * <strong>Note:</strong> Output from this program as well as from external classes (Medusa, Chinese
	 *                        Whispers etc.) is written to standard output/error. If only output of the
	 *                        <code>UnsuPosTagger</code> is of interest, the <code>PrintStream</code>s
	 *                        <code>sout</code> and <code>eout</code> can be specified and used.
	 * 
	 * @param fcorpusname     the name of the corpus file (one sentence per line).
	 * @param ignoreCase      controls if the class works case insensitive (<code>true</code>) or not
	 *                        (<code>false</code>).
	 * @param fconfname       the name of the configuration file.
	 * @param sout            an additional <code>PrintStream</code> for the standard output from this
	 *                        class only (can be <code>null</code>).
	 * @param eout            an additional <code>PrintStream</code> for the standard error from this class
	 *                        only (can be <code>null</code>). 
	 * @param newLineSequence the line separator for file output (should be <code>\n</code> or
	 *                        <code>\r\n</code>).
	 */
	public UnsuPosTagger(String fcorpusname, boolean ignoreCase, String fconfname,
			             PrintStream sout, PrintStream eout, String newLineSequence) throws Exception {
		
		/* Set informational task name and progress. */
		setTaskState("Initializing", "Read configuration file", 0);
		
		/* Test if corpus file exists. */
		fcorpus = new File(fcorpusname);
	    if (!fcorpus.exists() || !fcorpus.isFile()) {
	    	throw new Exception("Corpus file " + fcorpus.getPath() + " does not exist."); }
		
		/* Create temporary directory if it not existent. */
		File f = new File(System.getProperty("java.io.tmpdir"));
		if (!f.exists() && !f.mkdir()) {
			throw new Exception("Temporary directory does not exist and can not be created."); }
		printlnInfo("Use temporary directory " + f.getAbsolutePath() + ".");
	    
		/* Get directory of corpus file. */
		String dir = new File(fcorpus.getAbsolutePath()).getParent();
		
	    /* Create a directory for important files. */
		dirBasic = new File(dir + "/basic/");
		if ((!dirBasic.exists() && !dirBasic.mkdir()) || !dirBasic.isDirectory()) {
			throw new Exception("Directory " + dirBasic.getPath() + " can not be created."); }
		
		/* Create a directory for partitioning 1. */
		dirGraph1 = new File(dir + "/graph1/");
		if ((!dirGraph1.exists() && !dirGraph1.mkdir()) || !dirGraph1.isDirectory()) {
			throw new Exception("Directory " + dirGraph1.getPath() + " can not be created."); }
		
		/* Create a directory for partitioning 2. */
		dirGraph2 = new File(dir + "/graph2/");
		if ((!dirGraph2.exists() && !dirGraph2.mkdir()) || !dirGraph2.isDirectory()) {
			throw new Exception("Directory " + dirGraph2.getPath() + " can not be created."); }
		
		/* Create a directory for joined partitionings. */
		dirGraph3 = new File(dir + "/graph3/");
		if ((!dirGraph3.exists() && !dirGraph3.mkdir()) || !dirGraph3.isDirectory()) {
			throw new Exception("Directory " + dirGraph3.getPath() + " can not be created."); }
		
		/* Create a directory for the tagger model. */
		dirTmodel = new File(dir + "/tmodel/");
		if ((!dirTmodel.exists() && !dirTmodel.mkdir()) || !dirTmodel.isDirectory()) {
			throw new Exception("Directory " + dirTmodel.getPath() + " can not be created."); }
		
		/* Create a directory for the tagger model. */
		dirLog = new File(dir + "/log/");
		if ((!dirLog.exists() && !dirLog.mkdir()) || !dirLog.isDirectory()) {
			throw new Exception("Directory " + dirLog.getPath() + " can not be created."); }
		
		/* Create some files needed. */
		ftok      = new File(dirBasic, fcorpus.getName() + ".tok");
		fwn       = File.createTempFile(fcorpus.getName() + ".wn", null);
		fwf       = File.createTempFile(fcorpus.getName() + ".wf", null);
		File fiwn = File.createTempFile(fcorpus.getName() + ".internal.wn", null);
		File fiwf = File.createTempFile(fcorpus.getName() + ".internal.wf", null);
		fcosgraph = new File(dirBasic,  fcorpus.getName() + ".wcos_sorted_(2_0_1).edges");
		fsim12    = new File(dirBasic,  fcorpus.getName() + ".sim12.graph");
		fsim21    = new File(dirBasic,  fcorpus.getName() + ".sim21.graph");
		fTmodel   = new File(dirTmodel, fcorpus.getName() + ".tmodel");
		fcr       = new File(dirTmodel, fcorpus.getName() + ".cluster.read");
		
		/* Initialize Medusa (this must be done before the initialization of int variables?!?!?!). */
		try {
			ConfigurationContainer.setSentenceFileName(fcorpus.getPath());
			ConfigurationContainer.setTokenizedSentenceFileName(ftok.getPath());
			ConfigurationContainer.setExternalWordNumbersFileName(fwn.getPath());
			ConfigurationContainer.setExternalWordFrequenciesFileName(fwf.getPath());
			ConfigurationContainer.setWordNumbersFileName(fiwn.getPath());
			ConfigurationContainer.setWordFrequenciesFileName(fiwf.getPath());
			//ConfigurationContainer.setWordNumbersFileName(fwn.getPath());
			//ConfigurationContainer.setWordFrequenciesFileName(fwf.getPath());
		    ConfigurationContainer.setKnownWordNumbersFileName(dir + "/config/data/known-words.txt");
			ConfigurationContainer.setIDXFileName(File.createTempFile(fcorpus.getName() +
				 ".sidx", null).getPath());
			ConfigurationContainer.setPesistentHashFileName(File.createTempFile(fcorpus.getName() +
			     ".de.uni_leipzig.asv.medusa.loader.filter.IDXNeighbourhoodFilterImpl.hash", null).getPath());
			ConfigurationContainer.setWSWNFileName(File.createTempFile(fcorpus.getName() +
				 ".wswn", null).getPath());
			//ConfigurationContainer.setBOWWordFrequenciesFileName(fcorpus.getPath() + ".bbow");
			//ConfigurationContainer.setExportFileName(fcorpus.getPath() + ".eexport");
			
			init(); }
		catch (ConfigurationException e) {
			throw new Exception("Medusa configuration cannot be initialized.");	}
		
		totalTime = new Timer();
		totalTime.start();
		setTaskState("Initializing", "Read configuration file", 20);
		
		stopThread = false;
		
		/* Init the list of temp-file names. */
		tmpFileList = new ArrayList<String>();
		tmpFileList.add(fwn.getAbsolutePath());
		tmpFileList.add(fwf.getAbsolutePath());
		tmpFileList.add(fiwn.getAbsolutePath());
		tmpFileList.add(fiwf.getAbsolutePath());
		tmpFileList.add(ConfigurationContainer.getIDXFileName());
		tmpFileList.add(ConfigurationContainer.getPesistentHashFileName());
		tmpFileList.add(ConfigurationContainer.getExternalWordNumbersFileName());
		tmpFileList.add(ConfigurationContainer.getExternalWordFrequenciesFileName());
		tmpFileList.add(ConfigurationContainer.getWSWNFileName());
		
		/* Set additional standard output and standard error for this class only. */
		this.sout = sout;
		this.eout = eout;
		
		/* Case sensitivity. */
		this.ignoreCase = ignoreCase;
		
		/* Read config file. */
		try {
		    readConfig(fconfname); }
		catch (Exception e) {
			throw new Exception(e); }
		setTaskState(50);
		
		/* Set the new line character sequence. */
		if (newLineSequence == null) {
			NL = System.getProperty("line.separator"); }
		else {
			NL = newLineSequence; }
		if ((NL.compareTo("\n") != 0) && (NL.compareTo("\r\n") != 0)) {
			printlnError("New line sequence is neither '\\n' nor '\\r\\n'."); }
		
	    /* Initialization done. */
	    setTaskState(100);
		return; }
	
	/**
	 * This method does some preprocessing (corpus file tokenzing, creation of word and frequency list)
	 * and starts the unsupervised PoS-tagger.
	 */
	public void run() {
		Timer tm = new Timer();
		
		/* Determine the number of sentences in the corpus. */
		setTaskState("Initializing", "Determine the number of sentences in the corpus", 5);
		Nsmax = -1;
		try {
			printlnInfoBegin("Determine the number of sentences in the corpus ...", tm);
			BufferedReader fin = new BufferedReader(new FileReader(fcorpus.getPath()));
			int updateProgress = Math.max(1, (int) (fcorpus.length() / 14000));
			Nsmax = 0;
			String line;
			while ((line = fin.readLine()) != null) {
				if ((Nsmax % updateProgress) == 0) {
					increaseTaskProgress(1); }
				if (line.trim().length() > 0) {
					Nsmax++; } }
			setTaskState(100);
			printlnInfoEnd("... " + Nsmax + " sentences in the corpus", tm);
			
			/* Preprocessing (tokenizing and creation of word and frequency list). */
			if (needPreprocessing()) {
				/* Tokenize corpus file. */
				printlnInfoBegin("Start tokenizing ...", tm);
				doPreprocessing();
				printlnInfoEnd("... tokenizing done", tm); }
			
			if (!stopThread) {
				/* Create word number and frequency list. */
				printlnInfoBegin("Create word number and frequency list ... ", tm);
				createFrequencyList();
				printlnInfoEnd("... done", tm);
			
				/* Get number of tokens in the corpus. */
				setNumberOfWords();
			
				/* Start the unsupervised PoS-tagging. */
				start(); } }
		catch (Exception e) {
			e.printStackTrace();
			printlnError("Error while executing (" + e.getMessage() + ")"); }
		return;	}
	
	/**
	 * This method performs all steps of the unsupervised PoS-tagging.
	 */
	public void start() throws Exception {
		Timer tm = new Timer();
		
		/* Partitioning 1. */
		if (!stopThread && needPartitioning1()) {
			printlnInfoBegin("Start partitioning 1 (high frequent words)", tm);
			doPartitioning1();
			printlnInfoEnd("... partitioning 1 done", tm); }
		
		/* Partitioning 2. */
		if (!stopThread && needPartitioning2()) {
			printlnInfoBegin("Start partitioning 2 (medium and low frequent words)", tm);
			doPartitioning2();
			printlnInfoEnd("... partitioning 2 done", tm); }
		
		/* Join partitions. */
		if (!stopThread && needJoinPartitioning()) {
			printlnInfoBegin("Join partitionings", tm);
			joinPartitionings();
			printlnInfoEnd("... partitionings joined", tm); }
		
		/* Construct the tagger model. */
		if (!stopThread && needBuildLexicon()) {
			printlnInfoBegin("Construct lexicon ...", tm);
			buildLexicon();
			printlnInfoEnd("... lexicon constructed", tm); }
		
		/* Start Viterbi tagging. */
		if (!stopThread && needTagging()) {
			printlnInfoBegin("Start tagging ...", tm);
			doTagging();
			printlnInfoEnd("... tagging finished", tm); }
		
		cleanUp();
		totalTime.stop();
		printlnInfo("Total running time: " + totalTime.toString());
		return; }
	
    /**
	 * This method tokenizes a text file (one sentence per line). For tokenizing the
	 * <code>ImprovedWordTokenizer</code> class from the Medusa project is used.<p>
	 * 
	 * <strong>See also</strong>: de.uni_leipzig.asv.medusa<p>
	 *                            de.uni_leipzig.asv.medusa.input.ImprovedWordTokenizer <p>
	 *                            
	 * @throws Exception
	 */
	public void doPreprocessing() throws Exception {
		
		if (ftok.exists() && (doPreproc.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Tokenized corpus file " + ftok.getPath() + " already exists.");
			return; }
		
		/* Start tokenizing. */
		setTaskState("Preprocessing", "Tokenize corpus file", 0);
		
		/* Medusa tokenizing. */
		if (tokImpl.compareToIgnoreCase("medusa") == 0) {
			try {
				/* Initialize the tokenizer. */
				printInfoAlienOutputBegin("Medusa output");
				ImprovedWordTokenizer tok = new ImprovedWordTokenizer();
				tok.init();
				
		        /* Open input and output file. */
				BufferedReader fin  = new BufferedReader(new FileReader(fcorpus.getPath()));
				BufferedWriter fout = new BufferedWriter(new FileWriter(ftok.getPath())); 
			  
		        /* Read the input file and execute the tokenizer line by line. */
				boolean toLower = ignoreCase;
				String line, tokline;
				int n = 0, skipped = 0;
		        while ((line = fin.readLine()) != null) {
		        	n++;
		        	if (toLower) {
		        		tokline = tok.execute(line).trim().toLowerCase(); }
		        	else {
		        		tokline = tok.execute(line).trim(); }
		        	
		        	/* A number of tokens greater than 2 is necessary because of a Medusa bug in
		        	   input.AbstractInputImpl.getMWUTokens() (calling fct. input.SentenceToIdx2.processLine()) */ 
		        	if (tokline.split("\\s+").length <= 2) {
		        		skipped++; }
		        	else {
		        		fout.write(tokline + NL); }
			  	    if (n % (Nsmax / 100) == 0) { increaseTaskProgress(1); }
			  	    if (n % (Nsmax /  50) == 0) { printInfo(".");          } }
		        printInfoAlienOutputEnd();
		        printlnInfo("");
		        printlnInfo(skipped + " line(s) with less than three tokens skipped.");
		        setTaskState(100);
			    fin.close(); fout.close(); }
			catch (ConfigurationException e) {
				throw new Exception("Cannot initialize the tokenizer (" + e.getMessage() + ")."); }
			catch (Exception e) {
				throw new Exception("Could not tokenize corpus file " + fcorpus.getPath() + " ... " +
						            ftok.getPath() + " (" + e.getMessage() + ")."); } }
		
		/* Use standard tokenizing. */
		else if (tokImpl.compareToIgnoreCase("normal") == 0) {
			/* Open input and output file. */
			BufferedReader fin  = new BufferedReader(new FileReader(fcorpus.getPath()));
			BufferedWriter fout = new BufferedWriter(new FileWriter(ftok.getPath())); 
		  
	        /* Read the input file and execute the tokenizer line by line. */
			boolean toLower = ignoreCase;
			String line, tokline;
			int n = 0, skipped = 0;
	        while ((line = fin.readLine()) != null) {
	        	n++;
	        	if (toLower) {
	        		tokline = line.replaceAll("(\\+|\\-)?[ 0-9\\.\\,]+", "%N%").trim().toLowerCase(); }
	        	else {
	        		tokline = line.replaceAll("(\\+|\\-)?[ 0-9\\.\\,]+", "%N%").trim(); }
	        	
	        	/* A number of tokens greater than 2 is necessary because of a Medusa bug in
	        	   input.AbstractInputImpl.getMWUTokens() (calling fct. input.SentenceToIdx2.processLine()) */ 
	        	if (tokline.split("\\s+").length <= 2) {
	        		skipped++; }
	        	else {
	        		fout.write(tokline + NL); }
		  	    if (n % (Nsmax / 100) == 0) { increaseTaskProgress(1); }
		  	    if (n % (Nsmax /  50) == 0) { printInfo(".");          } }
	        printInfoAlienOutputEnd();
	        printlnInfo("");
	        printlnInfo(skipped + " line(s) with less than three tokens skipped.");
	        setTaskState(100);
		    fin.close(); fout.close(); }
		
		/* Unknown tokenizer. */
		else {
			throw new Exception("Unknown tokenizer '" + tokImpl + "'. Check configuration!"); }
		printlnInfo("");
		return; }
	
	/**
	 * This method determines the partitioning for high frequent words by means of a cosinus similarity
	 * graph.
	 * 
	 * @throws Exception
	 */
	public void doPartitioning1() throws Exception {
		Timer tm = new Timer();
		
		setStepName("Partitioning 1");
		try {
			fnodes   = new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.nodes");
			fedges   = new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.edges");
			fcluster = new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.cluster");
			
			/* Create cosinus similarity graph if necessary. */
			if (needCosGraph()) {
				createCosGraph(); }
			
			/* Create graph 1 from cosine similarity graph. */
			if (needGraph1()) {
				createGraph1(fedges.getPath(), fnodes.getPath()); }
			
			/* Run CW clustering and move cluster file the ./graph1 directory. */
			if (needClustering1()) {
				printlnInfoBegin("Clustering of graph (high frequent words) ...", tm);
				runCWClustering(fnodes.getPath(), fedges.getPath(), fcluster.getPath());
				//moveFile(fcluster.getPath(), new File(dirGraph1.getPath(), fcluster.getName()).getPath());
				printlnInfoEnd("... clustering done", tm); } }
		catch (Exception e) {
			throw new Exception(e); }
		return;	}
	
	/**
	 * This method creates the partitioning for medium and low frequent words.
	 * 
	 * @throws Exception
	 */
	public void doPartitioning2() throws Exception {
		Timer tm = new Timer();
		
		setStepName("Partitioning 2");
		try {
			fnodes   = new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.nodes");
			fedges   = new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.edges");
			fcluster = new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.cluster");
			
			/* Create similarity graph. */
			if (needGraph2()) {
				createGraph2(fedges.getPath(), fnodes.getPath()); }
			
			/* Run CW clustering and move cluster file the ./graph1 directory. */
			if (needClustering2()) {
				printlnInfoBegin("Clustering of graph (medium and low frequent words) ...", tm);
				runCWClustering(fnodes.getPath(), fedges.getPath(), fcluster.getPath());
				//moveFile(fcluster.getPath(), new File(dirGraph2.getPath(), fcluster.getName()).getPath());
				printlnInfoEnd("... clustering done", tm); } }
		catch (Exception e) {
			throw new Exception(e); }
		return;	}
	
	/**
	 * This method combines the first and second partitioning (high and medium/low frequent words,
	 * respectively).
	 * 
	 * @throws Exception
	 */
	public void joinPartitionings() throws Exception {
		Timer tm = new Timer();
		
		setStepName("Combine partitionings");
		try {
			String cl1FileName, cl2FileName;
			cl1FileName = new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.cluster").getPath();
			cl2FileName = new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.cluster").getPath();
			
			fnodes   = new File(dirGraph3.getPath(), fcorpus.getName() + ".graph3.nodes");
			fedges   = new File(dirGraph3.getPath(), fcorpus.getName() + ".graph3.edges");
			fcluster = new File(dirGraph3.getPath(), fcorpus.getName() + ".graph3.cluster");
			createGraph3(cl1FileName, cl2FileName, fedges.getPath(), fnodes.getPath()); }
		catch (Exception e) {
			throw new Exception(e); }
		
		/* If graph3 is not empty. */
		if (!stopThread) {
			
			/* Create temporary word list. */
			printlnInfoBegin("Clustering of graph (joined partitionings) ...", tm);
			runCWClustering(fnodes.getPath(), fedges.getPath(), fcluster.getPath());
			//fcluster.renameTo(new File(dirGraph3.getPath(), fcluster.getName()));
	        printlnInfoEnd("... clustering done.", tm);
	        
	        printlnInfoBegin("Create final partitioning ...", tm);
	        String fcn = fcorpus.getName();
	        resolveCluster(new File(dirGraph1, fcn + ".graph1.cluster.read").getPath(),
	        		       new File(dirGraph2, fcn + ".graph2.cluster.read").getPath(),
	        		       new File(dirGraph3, fcn + ".graph3.cluster.read").getPath());
	        /*resolveCluster(fcorpus.getPath() + ".graph1.cluster.read",
	        		       fcorpus.getPath() + ".graph2.cluster.read",
	 		               fcorpus.getPath() + ".graph3.cluster.read");*/
	        printlnInfoEnd("... final partitioning created", tm); }
		return; }
	
	/**
	 * This method constructs a lexicon from trigram statistics.
	 * 
	 * @throws Exception
	 */
	public void buildLexicon() throws Exception {
		BufferedReader fin;
		BufferedWriter fout;
		String[] tagId2tag;
		HashMap<String, String>  w2tag;
		HashMap<String, Integer> tag2tagId, w2tagId, wordsToWrite, freqList;
		String line;
		int NtagId = -1;
		
		/***********************************************************************************
		 * Variables:
		 *   w2tag    :    maps word to tag name (cluster number).  HashMap<String, String>
		 *   tag2tagId:    maps tag name to tag ID.                 HashMap<String, Integer>
		 *   tagId2tag:    maps tag ID to tag name.                 String[]
		 *   freqList :    maps word to word frequency.             HashMap<String, Integer>
		 *   w2tagId  :    maps word to tag ID.                     HashMap<String, Integer>
		 ***********************************************************************************/
		
		setTaskState("Create tagger model", "Write lexicon", 0);
		try {
			/* Determine the number of clusters/tags (plus for for tag <BPOS>). */
			NtagId = getMaxTagNumber() + 1;
		
			/* Create a list of words and corresponding tags. */
			fin = new BufferedReader(new FileReader(fcr));
			w2tag  = new HashMap<String, String>();
			w2tag.put("<BPOS>", "-");
			tagId2tag    = new String[NtagId];
			tag2tagId    = new HashMap<String, Integer>();
			w2tagId      = new HashMap<String, Integer>();
			wordsToWrite = new HashMap<String, Integer>();
			int id = 0;
			while ((line = fin.readLine()) != null) {
				String[] cluster = line.split("\t");
				tagId2tag[id] = cluster[0];
				tag2tagId.put(cluster[0], id);
				for (String word : cluster[2].split(", ")) {
					w2tag.put(word, cluster[0]);
					w2tagId.put(word, id);
					wordsToWrite.put(word, id); }
				id++; }
			fin.close();
			setTaskState(15);
			
			/* Check for dummy tag '-'. */
			if (tag2tagId.get("-") == null) {
				tagId2tag[id] = "-";
				tag2tagId.put("-", id++); }
			else {
				throw new Exception("Error while building tag list: tag '-' is already in the tag list."); }
			
			/* Test if n = NtagId */
			if (id != NtagId) {
				throw new Exception("Error while building tag list from *.cluster.read"); }
		
			/* Write the *.taglist file. */
			fout = new BufferedWriter(new FileWriter(fTmodel.getPath() + ".taglist"));
			for (int i = 0; i < NtagId; i++) {
				fout.write(tagId2tag[i] + NL); }
			fout.close();
			setTaskState(30);
			
			/* Get first Ntop+ edges from cosinus similarity graph. */
			/* String fcgNtopname = fTmodel.getPath() + "." + Ntop + ".edges"; */
			String fcgNtopname = File.createTempFile(fcorpus.getName() + "." +
					                                 Ntop + ".edges", null).getAbsolutePath();
			tmpFileList.add(fcgNtopname);
			
			int maxNtopwno = writeFirstNodes(fcosgraph.getPath(), fcgNtopname, Ntop);
			setTaskState(45);

			/* Read a word list for the cosine similarity graph which maps word numbers to words. */
			fin  = new BufferedReader(new FileReader(fwn.getPath()));
			String[] wordList = new String[maxNtopwno + 1];
			while ((line = fin.readLine()) != null) {
				String[] word = line.split("\t");
				int wno = Integer.parseInt(word[0]);
				if (wno > maxNtopwno) {
					break; }
				wordList[wno] = word[1]; }
			fin.close();
			setTaskState(60);
			
			/* Get frequency list (word |-> frequency). */
			int maxwno = Math.max(maxNtopwno , getMaxWordNumberFromClusterFiles());
			freqList = readFreqList(1, maxwno);
			printlnInfo("Biggest word number: " + maxwno);
			
			/* Write the *.lexicon file with ambiguous words. */
			fin  = new BufferedReader(new FileReader(fcgNtopname));
			fout = new BufferedWriter(new FileWriter(fTmodel.getPath() + ".lexicon"));
			HashMap<Integer, Integer>     ambigSum  = new HashMap<Integer, Integer>();
			HashMap<IntegerPair, Integer> ambigProb = new HashMap<IntegerPair, Integer>();
			Integer freq, tId;
			
			/************************************************************************
			 * Read all edges from cosinus similarity graph.
			 * 1) Write all non-ambiguous words to output file (with corresponding
			 *    membership value equal to 1).
			 * 2) Write all ambiguous words to output file with membership values
			 *    according to connectivity in cosinus similarity graph.
			 ************************************************************************/
			while ((line = fin.readLine()) != null) {
				String[] edge = line.split("\t");
				String word0 = wordList[Integer.parseInt(edge[0])];
				String word1 = wordList[Integer.parseInt(edge[1])];
				
				/* If first node (word) is member of a cluster ... */
				if (w2tag.get(word0) != null) {
					
					/* If not already written to lexicon. */
					if ((tId = wordsToWrite.remove(word0)) != null) {
						
						/* Write with corresponding membership value 1. */
						if ((freq = freqList.get(word0)) == null) {
							freq = new Integer(1); }
						fout.write(freq.intValue() + "\t" + word0);
						for (int tagId = 0; tagId < NtagId; tagId++) {
							if (tId.intValue() == tagId) {
								fout.write("\t1"); }
							else {
								fout.write("\t0"); } }
						fout.write(NL); }
					
					/* Second node (word) is member of a cluster. */
					if (w2tag.get(word1) != null) {
						
						/* If not already written to lexicon. */
						if ((tId = wordsToWrite.remove(word1)) != null) {
							
							/* Write with corresponding membership value 1. */
							if ((freq = freqList.get(word1)) == null) {
								freq = new Integer(1); }
							fout.write(freq.intValue() + "\t" + word1);
							for (int tagId = 0; tagId < NtagId; tagId++) {
								if (tId.intValue() == tagId) {
									fout.write("\t1"); }
								else {
									fout.write("\t0"); } }
							fout.write(NL); } }
					
					/* Second node (word1) is not a member of any cluster (i.e. ambiguous) but *
					 * connected to a node (word0) which itself is a member of a cluster.      */
					else {
						/* Sum the overall connection strength of edge[1] (word1). */
						Integer sum;
						if ((sum = ambigSum.get(Integer.parseInt(edge[1]))) != null) {
							sum += Integer.parseInt(edge[2]); }
						else {
							sum  = Integer.parseInt(edge[2]); }
						ambigSum.put(Integer.parseInt(edge[1]), sum);
						
						/* Sum weights of connections to the corresponding cluster (tag of word0). */
						IntegerPair wc = new IntegerPair(Integer.parseInt(edge[1]),
								                         w2tagId.get(word0).intValue());
						Integer weight;
						if ((weight = ambigProb.get(wc)) != null) {
							ambigProb.put(wc, Integer.parseInt(edge[2]) + weight.intValue()); }
						else {
							ambigProb.put(wc, Integer.parseInt(edge[2])); } } } }
			fin.close();
			setTaskState(90);

			
			/* Write remaining cluster members. */
			for (String word : wordsToWrite.keySet()) {
				tId = wordsToWrite.get(word);
				
				/* Write with corresponding membership value 1. */
				if ((freq = freqList.get(word)) == null) {
					freq = new Integer(1); }
				fout.write(freq.intValue() + "\t" + word);
				for (int tagId = 0; tagId < NtagId; tagId++) {
					if (tId.intValue() == tagId) {
						fout.write("\t1"); }
					else {
						fout.write("\t0"); } }
				fout.write(NL); }
			wordsToWrite.clear(); wordsToWrite = null;
			setTaskState(95);
			
			/* Write all ambiguous words to the lexicon. */
			System.out.println("Write " + ambigSum.size() + " ambiguous words ...");
			for (int wordno : ambigSum.keySet()) {
				String word = wordList[wordno];
				
				/* Get node number and 'connected' tag. */
				fout.write(freqList.get(word).intValue() + "\t" + word);
				IntegerPair wc = new IntegerPair(wordno, -1);
				for (int tagId = 0; tagId < NtagId; tagId++) {
					
					wc.e2 = tagId;
					
					/* If word is connected to the actual cluster/tag. */
					Integer weight;
					if ((weight = ambigProb.get(wc)) == null) {
						fout.write("\t0"); }
					else {
						double prob = ((double) weight.intValue()) / ((double) ambigSum.get(wordno).intValue());
						fout.write("\t" + prob); } }
				fout.write(NL); }
			fout.close();
			setTaskState(100);

			/* Count trigrams. */
			setTaskState("Count trigrams", 0);
			fin = new BufferedReader(new FileReader(ftok.getPath()));
			HashMap<String, Integer> trigrams = new HashMap<String, Integer>();
			int Ntrigrams = 0, Nknowntrigrams = 0;
			int update = Math.max(1, (int) Math.round(Nsmax / 100.0));
			int l = 0;
			while ((line = fin.readLine()) != null) {
				if ((l++ % update) == 0) {
					increaseTaskProgress(1); }
				line = "<BPOS> <BPOS> " + line;
				String[] tokens = line.trim().split("\\s+");
				if (tokens.length >= 3) {
					String[]  tagNames = new String[tokens.length];
					String    tagName  = null;
					boolean[] known    = new boolean[tokens.length];
					for (int i = 0; i < tokens.length; i++) {
						tagNames[i] = "";
						known[i]    = false;
						if ((tagName = w2tag.get(tokens[i])) != null) {
							tagNames[i] = tagName;
							known[i]    = true; } }
					
					for (int i = 2; i < tokens.length; i++) {
						Ntrigrams++;
						if (known[i-2] && known[i-1] && known[i]) {
							Nknowntrigrams++;
							String sequence = tagNames[i-2] + "|" + tagNames[i-1] + "|" + tagNames[i];
							if ((freq = trigrams.get(sequence)) == null) {
								trigrams.put(sequence, 1); }
							else {
								trigrams.put(sequence, freq.intValue() + 1); } } }
					tagNames = null; known = null; } }
			fin.close();
			setTaskState(100);
			
			int coverage = (int) Math.round(100.0 * (double) Nknowntrigrams / (double) (Ntrigrams + 1));
			printlnInfo("Known trigrams: " + coverage +
			                   "% (" + Nknowntrigrams + " / " + Ntrigrams + ").");
			
			/* Write the *.transitions file. */
			setTaskState("Write transition file", 0);
			fout = new BufferedWriter(new FileWriter(fTmodel.getPath() + ".transitions"));
            update = Math.max(1, (int) Math.round(NtagId / 25.0));
			Integer n;
			for (int i = 0; i < NtagId; i++) {
				if ((i % update) == 0) {
					increaseTaskProgress(4); }
				for (int j = 0; j < NtagId; j++) {
					int[] values = new int[NtagId];
					boolean exists = false;
					int sum = 0;
					for (int k = 0; k < NtagId; k++) {
						String sequence = tagId2tag[i] + "|" + tagId2tag[j] + "|" + tagId2tag[k];
						values[k] = 0;
						if ((n = trigrams.get(sequence)) != null) {
							exists     = true;
							values[k]  = n.intValue();
							sum       += n.intValue(); } }
					if (exists) {
						fout.write(tagId2tag[i] + "|" + tagId2tag[j]);
						for (int k = 0; k < NtagId; k++) {
							double p = ((double) values[k]) / ((double) sum);
							
							/* Write 0 instead of 0.0 due to Viterbi tagger 'requirements'. */
							if (p < 1.0e-9) {
								fout.write("\t0"); }
							else {
								fout.write("\t" + p); } }
						fout.write(NL); } } }
			fout.close();
			setTaskState(100); }
		catch (Exception e) {
			throw new Exception(e); }
		
		writeModelFile();
		
		return; }
	
	/**
	 * This method writes the model file for the Viterbi tagger.
	 *  
	 * @throws Exception
	 */
	private void writeModelFile() throws Exception {
		try {
			BufferedWriter fout = new BufferedWriter(new FileWriter(fTmodel.getPath() + ".model"));
			fout.write("taglist=" + fTmodel.getPath() + ".taglist" + NL);
			fout.write("secondTag=false" + NL);
			if (ConfigurationContainer.getCategory("de.uni_leipzig.asv.medusa.input.ImprovedWordTokenizer").getBooleanProperty("boolReplaceNumbers", false)) {
				fout.write("ReplaceNumbers=true" + NL); }
			else {
				fout.write("ReplaceNumbers=false" + NL); }
			fout.write("transitions=" + fTmodel.getPath() + ".transitions" + NL);
			fout.write("lexicon=" + fTmodel.getPath() + ".lexicon" + NL);
			fout.close(); }
		catch (Exception e) {
			throw new Exception(e); }
		return; }
	
	/**
	 * This method calls the <code>ViterbiTagger</code> for the tokenized corpus file with the tagger model
	 * from <code>*.tmodel</code>. 
	 */
	public void doTagging() {
		setTaskState("Start Viterbi tagger", "Tagging", 5);
		String[] args = { fTmodel.getPath(), ftok.getPath() };
		viterbitagger.Main.main(args);
		setTaskState(100);
		return; }
	
	/**
	 * This method returns the number of sentences in the corpus.
	 * 
	 * @return int the number of sentences in the tokenized corpus file.
	 */
	public int getNumberOfSentences() {
		return Nsmax; }
	
	/**
	 * This method returns the greatest word number (i.e. the number of words).
	 *  
	 * @return int the greatest word number
	 */
	public int getNumberOfWords() {
		return Nwmax; }
	
	/**
	 * This method prints some information about the Medusa configuration on standard output.
	 */
	public void printConfigInfo() {
		printlnInfo("");
		printlnInfo("General configuration");
		printlnInfo("  Temp. directory: " + System.getProperty("java.io.tmpdir"));
		if (deleteTempFiles) {
			printlnInfo("  Temporary files are deleted after execution."); }
		printlnInfo("UnsuPoS configuration");
		printlnInfo("  FEAT         = " + Nwfeat);
		printlnInfo("  TARG         = " + Nwtarg);
		printlnInfo("  CWTARG       = " + NCWtarg);
		printlnInfo("  COS_THRESH   = " + Tcos);
		printlnInfo("  TOPADD       = " + Ntop);
		printlnInfo("  NB_THRESH    = " + Tnb);
		printlnInfo("  NB_MAX       = " + Nnbmax);
		printlnInfo("  CONF_OVERLAP = " + Noverlap);
		printlnInfo("  SING_ADD     = " + Nsingadd);
		printlnInfo("  BEHEAD       = " + Nbehead);
		printlnInfo("Medusa configuration");
		printlnInfo("  CONTROL_FLOW_IMPL");
		printlnInfo("    " + getGlobalProperty("CONTROL_FLOW_IMPL"));
		printlnInfo("  SENTENCE_FILE_NAME");
		printlnInfo("    " + getGlobalProperty("SENTENCE_FILE_NAME"));
		printlnInfo("  PARSER_FILTER_IMPL");
		printlnInfo("    " + getGlobalProperty("PARSER_FILTER_IMPL"));
		printlnInfo("  -----------------------------------------------------------------");
		printlnInfo("  Tokenized File: "  + ConfigurationContainer.getTokenizedSentenceFileName());
		printlnInfo("  Word numbers  : "  + ConfigurationContainer.getWordNumbersFileName());
		printlnInfo("  Word numbers  : "  + ConfigurationContainer.getWordNumbersCompleteName());
		printlnInfo("  IDX file      : "  + ConfigurationContainer.getIDXFileName());
		printlnInfo("  Bow file      : "  + ConfigurationContainer.getBOWWordFrequenciesFileName());
		return; }
	
	/**
	 * This method returns the description of the actual step (preprocessing, partitioning 1 ...).
	 * One step normally consists of several tasks (see <code>getActualTaskName()</code>).
	 * 
	 * @return a description of the actual step.
	 */
	public String getActualStepName() {
		return stepName; }
	
	/**
	 * This method returns the description of the actual task.
	 * 
	 * @return a description of the actual task.
	 */
	public String getActualTaskName() {
		return taskName; }
	
	/**
	 * This method returns the progress of the actual task.
	 * 
	 * @return the progress of the actual task.
	 */
	public int getActualTaskProgress() {
		return taskProgress; }
	
	/**
     * Reads the configuration file specified by the fname argument. This method
     * returns true on success, otherwise false.
     * 
     * @param  fname the configuration file name 
     */
    public void readConfig(String fname) throws Exception {
    	String p;
     	try {
    	    FileInputStream in = new FileInputStream(fname);
    	    settings.load(in); }
    	catch (IOException  e) {
    		throw new Exception("Cannot read parameters from " + fname + "."); }
    	
    	if (((p = settings.getProperty("DELTEMPFILES")) != null) && (p.compareToIgnoreCase("true") != 0)) {
    		deleteTempFiles = false; }
    	else {
    		deleteTempFiles = true; }
    	
    	/* Read the number of feature words. */
    	if ((p = settings.getProperty("FEAT")) != null) {
    		Nwfeat = Integer.parseInt(p); }
    	else {
    		Nwfeat = 200;
    		printlnError("Parameter FEAT (number of feature words) not found in " + fname + ".");
    		printlnError("  FEAT = " + Nwfeat + " will be used instead ..."); }
    	
    	/* Read the number of target words. */
    	if ((p = settings.getProperty("TARG")) != null) {
    		Nwtarg = Integer.parseInt(p); }
    	else {
    		Nwtarg = 10000;
    		printlnError("Parameter TARG (number of target words) not found in " + fname + ".");
    		printlnError("  TARG = " + Nwtarg + " will be used instead ..."); }
    	
    	if (Nwfeat > Nwtarg) {
    		throw new Exception("Number of feature words is greater than number of target words" +
    				            " (FEAT > TARG)."); }
    	
    	/* Read the number of feature words for CW. */
    	if ((p = settings.getProperty("CWTARG")) != null) {
    		NCWtarg = Integer.parseInt(p); }
    	else {
    		NCWtarg = 5000;
    		printlnError("Parameter CWTARG (number of feature words for CW) not found in " +
    				           fname + ".");
    		printlnError("  CWTARG = " + NCWtarg + " will be used instead ..."); }
    	
    	/* Read the number of words considered for the lexicon. */
    	if ((p = settings.getProperty("TOPADD")) != null) {
    		Ntop = Integer.parseInt(p); }
    	else {
    		Ntop = 9500;
    		printlnError("Parameter TOPADD not found in " +
    				           fname + ".");
    		printlnError("  TOPADD = " + Ntop + " will be used instead ..."); }
    	
    	/* Read the threshold for cosinus similarity between context vectors. */
    	if ((p = settings.getProperty("COS_THRESH")) != null) {
    		Tcos = Double.parseDouble(p); }
    	else {
    		Tcos = 0.66;
    		printlnError("Parameter COS_THRESH not found in " + fname + ".");
    		printlnError("  COS_THRESH = " + Tcos + " will be used instead ..."); }
    	
    	/* Read the threshold S_THRESH. */
    	/*if ((p = settings.getProperty("S_THRESH")) != null) {
    		Tsig = Double.parseDouble(p); }
    	else {
    		Tsig = 2.71;
    		printlnError("Parameter S_THRESH not found in " + fname + ".");
    		printlnError("  S_THRESH = " + Tsig + " will be used instead ..."); }*/
    	
    	/* Read the threshold NB_THRESH. */
    	if ((p = settings.getProperty("NB_THRESH")) != null) {
    		Tnb = Integer.parseInt(p); }
    	else {
    		Tnb = 4;
    		printlnError("Parameter NB_THRESH not found in " + fname + ".");
    		printlnError("  NB_THRESH = " + Tnb + " will be used instead ..."); }
    	
    	/* Read the threshold NB_THRESH. */
    	if ((p = settings.getProperty("NB_MAX")) != null) {
    		Nnbmax = Integer.parseInt(p); }
    	else {
    		Nnbmax = 500;
    		printlnError("Parameter NB_MAX not found in " + fname + ".");
    		printlnError("  NB_MAX = " + Nnbmax + " will be used instead ..."); }
    	
    	/* Read the threshold BEHEAD. */
    	if ((p = settings.getProperty("BEHEAD")) != null) {
    		Nbehead = Integer.parseInt(p); }
    	else {
    		Nbehead = 2000;
    		printlnError("Parameter BEHEAD not found in " + fname + ".");
    		printlnError("  BEHEAD = " + Nbehead + " will be used instead ..."); }
    	
    	/* Read the the value for minimal overlap CONF_OVERLAP. */
    	if ((p = settings.getProperty("CONF_OVERLAP")) != null) {
    		Noverlap = Integer.parseInt(p); }
    	else {
    		Noverlap = 2;
    		printlnError("Parameter CONF_OVERLAP not found in " + fname + ".");
    		printlnError("  CONF_OVERLAP = " + Noverlap + " will be used instead ..."); }
    	
    	/* Read the the number of singletons SING_ADD to add. */
    	if ((p = settings.getProperty("SING_ADD")) != null) {
    		Nsingadd = Integer.parseInt(p); }
    	else {
    		Nsingadd = 200;
    		printlnError("Parameter SING_ADD not found in " + fname + ".");
    		printlnError("  SING_ADD = " + Nsingadd + " will be used instead ..."); }
    	
    	/* Control steps to perform. */
    	if ((doPreproc   = settings.getProperty("PREPROC"))   == null) {
    		doPreproc    = "force"; }
    	if ((tokImpl     = settings.getProperty("TOKENIZER")) == null) {
    		tokImpl      = "medusa"; }
    	if ((doPart1     = settings.getProperty("PART1"))     == null) {
    		doPart1      = "force"; }
    	if ((doPart2     = settings.getProperty("PART2"))     == null) {
    		doPart2      = "force"; }
    	if ((doJoinParts = settings.getProperty("JOINPARTS")) == null) {
    		doJoinParts  = "force"; }
    	if ((doBuildLex  = settings.getProperty("BUILDLEX"))  == null) {
    		doBuildLex   = "force"; }
    	if ((doTagg      = settings.getProperty("TAGGING"))   == null) {
    		doTagg       = "force"; }
    	return; }
    
    
    /**
     * This method determines the number of words (tokens) in the tokenized corpus file.
     * 
     * @throws Exception
     */
    private void setNumberOfWords() throws Exception {
    	try {
    		BufferedReader fin  = new BufferedReader(new FileReader(fwn.getPath()));
    		Nwmax = -1;
    		long n = 0;
    		while (fin.readLine() != null) {
    			n++; }
    		fin.close();
		
    		/* Test if greatest word number is an integer (it should be). */ 
    		if (n > (long) Integer.MAX_VALUE) {
    			throw new Exception("Number of words is greater than " + Integer.MAX_VALUE +
    			                    " (Integer.MAX_VALUE)."); }
    		else {
    			Nwmax = (int) n; } }
    	catch (FileNotFoundException e) {
			throw new Exception("Word list file not found (" + fwn.getPath() + ")."); }
    	return; }
    
    /**
     * This method sets the <code>stopThread</code> variable to true to signal that the thread should be stopped as
     * soon as possible.
     * 
     * @param msg a message string to be printed.
     */
    private void stopHere(String msg) {
    	printlnInfo(msg);
    	stopThread = true;
    	return; }
    
    /**
     * This method creates the cosine similarity graph. For every target word the context -2, -1, 1, 2
	 * is examined.
	 * 
     * @throws Exception
     */
    private void createCosGraph() throws Exception {
    	Timer tm = new Timer();
    	
		setTaskState("Create context vectors", 0);
		
		if (true) {
			/* Initialize matrix V of context vectors. */
			printlnInfoBegin("Initialize matrix of context vectors ...", tm);
			V = new int[Nwtarg + 1][4 * Nwfeat];
			for (int i = 1; i <= Nwtarg; i++) {
				for (int j = 0; j < 4 * Nwfeat; j++) {
					V[i][j] = 0; } }
			printlnInfoEnd("... done", tm);
		
			/* Read word list into Hashmap. */
			printlnInfoBegin("Read wordlist into hashmap ...", tm);
			HashMap<String,  Integer> wordlist = readWordList(1, Nwtarg);
			printlnInfoEnd("... done", tm);

			/* Read tokenized corpus and count feature words in <win>-neighbourhood. */
			printlnInfoBegin("Read corpus and create context vectors ...", tm);
			try {
				BufferedReader fin = new BufferedReader(new FileReader(ftok.getPath()));
				String line;
				String[] words;
				int offset, n = 0;
				while ((line = fin.readLine()) != null) {
					n++;
					if (n % (Nsmax / 100) == 0) { increaseTaskProgress(1); }
					if (n % (Nsmax /  50) == 0) { printInfo(".");          }
					
					/* Tokenize sentence. */
					words = line.split("\\s+");
					
					/* Investigate context for all words. */
					for (int i = 0; i < words.length; i++) {
						
						/* If the actual word is a target word. */
						Integer featID, wordID = Nwtarg + 1;
						if (((wordID = wordlist.get(words[i])) != null) && (wordID.intValue() <= Nwtarg)) {
							
							/* Scan context of word at position i in the actual sentence. */
							for (int p = 1; p <= 2; p++) {
								
								/* Scan left context. */
								if (i - p >= 0) {
									if (((featID = wordlist.get(words[i - p])) != null) &&
										(featID.intValue() <= Nwfeat)) {
										offset = 4 * (featID.intValue() - 1) + 2 * p - 2;
										V[wordID][offset] = V[wordID][offset] + 1; } }
								
								/* Scan right context. */
								if (i + p < words.length) {
									if (((featID = wordlist.get(words[i + p])) != null) &&
										(featID.intValue() <= Nwfeat)) {
										offset = 4 * (featID.intValue() - 1) + 2 * p - 1;
										V[wordID][offset] = V[wordID][offset] + 1; } } } } } }
				
				/* Context vectors created. */
				setTaskState(100); }
			catch (FileNotFoundException e) {
				throw new Exception("Tokenized corpus file not found (" + ftok.getPath() + ")."); }
			printlnInfo("");
			printlnInfoEnd("... context vectors created", tm);
			
			/* Word list no longer needed. */
			wordlist.clear(); wordlist = null;
			
			/* Calculate the Euclidian norm of every context vector. */
			printlnInfoBegin("Calculate length of context vectors ...", tm);
			vnorm = new double[Nwtarg + 1];
			double vij;
			for (int i = 1; i <= Nwtarg; i++) {
				if (i % (Nwtarg / 50) == 0) { printInfo("."); }
				vnorm[i] = 0.0;
				for (int j = 0; j < 4 * Nwfeat; j++) {
					vij = (double) V[i][j];
					vnorm[i] += vij * vij; }
				vnorm[i] = Math.sqrt(vnorm[i]); }
			printlnInfo("");
			printlnInfoEnd("... length of context vectors calculated", tm);
			
			/* Calculate the cosinus of the angle between context vectors. */
			int Ncpu = Math.min(16, Runtime.getRuntime().availableProcessors());
			setTaskState("Calculate cosine similarity (" + Ncpu + " thread(s))", 0);
			printlnInfoBegin("Calculate cosine similarity graph (" + Ncpu + " thread(s)) ...", tm);
			
			/* Start some threads. */
			Thread[]   p     = new Thread[Ncpu];
			MyCosRun[] mycos = new MyCosRun[Ncpu];
			
			File tmpFileEdges = File.createTempFile(fcorpus.getName() + ".wcos.edges", null);
			BufferedWriter fout = new BufferedWriter(new FileWriter(tmpFileEdges));
			for (int pno = 0; pno < Ncpu; pno++) {
				mycos[pno] = new MyCosRun(pno + 1, Ncpu, fout);
				p[pno]     = new Thread(mycos[pno]);
				p[pno].start(); }
			
			/* Check progress. */
			boolean done = false;
			while (!done) {
				Thread.sleep(500);
				int psum = 0;
				boolean anyAlive = false;
				for (int pno = 0; pno < Ncpu; pno++) {
					psum += mycos[pno].taskProgress;
					anyAlive = anyAlive || p[pno].isAlive(); }
				done = !anyAlive;
				setTaskState(psum / Ncpu); }
			
			/* Wait for termination. */
			for (int pno = 0; pno < Ncpu; pno++) {
				p[pno].join(); }
			fout.close();
			printlnInfo("");
			printlnInfoEnd("... graph calculated", tm);
			
			/* Sort list of edges according to weight value: order={ 0 2 1 }, types={ i i I }. */
			setTaskState("Sort edge list and use first " + NCWtarg + "+ nodes", 20);
			printlnInfoBegin("Sort edge list ...", tm);
			int[]  sOrder = {2, 0, 1};
			char[] sType  = {'i', 'i', 'I'};
			FileSort s = new FileSort("\t", sOrder, sType);
			s.sort(tmpFileEdges.getAbsolutePath(), fcosgraph.getPath());
			//deleteFileIfExists(tmpFileEdges.getAbsolutePath());
			tmpFileList.add(tmpFileEdges.getAbsolutePath());
			printlnInfoEnd("... edge list sorted", tm); }
		setTaskState(90);
    	return; }
    
	/**
	 * This method creates the graph from the cosine similarity graph (for partitioning 1).
	 * 
	 * @throws Exception
	 */
	private void createGraph1(String edgeFileName, String nodeFileName) throws Exception {
		setTaskState("Write node and edge list for graph 1", 10);
		
		/* Get first NCWtarg+ edges. */
		int maxNodeId = writeFirstNodes(fcosgraph.getPath(), edgeFileName, NCWtarg);
		printlnInfo(maxNodeId + " biggest word number (" + edgeFileName + ")");
		setTaskState(60);
		
		/* Write node list (first Nwtarg words). */
		writeWordList(nodeFileName, 1, maxNodeId);
		setTaskState(100);
		return; }
	
	/**
	 * This method constructs the graph for the partitioning 2.
	 */
	private void createGraph2(String edgeFileName, String nodeFileName) throws Exception {
		Timer tm = new Timer();
		BufferedReader fin;
		File fnbsymTmp = File.createTempFile(fcorpus.getName() + ".nbsym.graph", null);
		File fnb12Tmp  = File.createTempFile(fcorpus.getName() + ".nb12.graph", null);
		File fnb21Tmp  = File.createTempFile(fcorpus.getName() + ".nb21.graph", null);
		
		tmpFileList.add(fnbsymTmp.getAbsolutePath());
		tmpFileList.add(fnb12Tmp.getAbsolutePath());
		tmpFileList.add(fnb21Tmp.getAbsolutePath());
		
		try {
			/*
			if (fsim12.exists() && fsim21.exists() && (doPart2.compareToIgnoreCase("force") != 0)) {
				printlnInfo("  Similarity files already exist (skip creation)."); }
			else {*/
			
			/* Use nb test graph .nbgraph.test if existent. */
			String hashFileName;
			File fnbgraph = new File(fcorpus.getAbsolutePath() + ".nbgraph.test");
			if (fnbgraph.exists() && fnbgraph.isFile()) {
				hashFileName = fnbgraph.getAbsolutePath();
				printlnInfo("Use existing nb graph: " + fnbgraph.getPath() + "."); }
				
			/* Calculate cooccurrences. */
			else {
				setTaskState("Calculate cooccurrences", 0);
				printlnInfoBegin("Calculate cooccurrences ...", tm);
				MyCooccRun cAlg = new MyCooccRun();
				cAlg.start();
				cAlg = null;
				printlnInfoEnd("... cooccurrences calculated", tm);
				hashFileName = ConfigurationContainer.getPesistentHashFileName() + ".lgl2.expo"; }
				
			/* Read calculated cooccurrences and construct graph 2. */
			setTaskState("Calculate similarity", 0);
			printlnInfoBegin("Calculate similarity ...", tm);
			BufferedWriter[] fouts = new BufferedWriter[3];
			fouts[0] = new BufferedWriter(new FileWriter(fnbsymTmp));
			fouts[1] = new BufferedWriter(new FileWriter(fnb12Tmp));
			fouts[2] = new BufferedWriter(new FileWriter(fnb21Tmp));
				
			fin = new BufferedReader(new FileReader(hashFileName));
			String line;
			int sig;
			while ((line = fin.readLine()) != null) {
				String[] coocc = line.split("\t");
				sig = (int) (100.0 * Double.parseDouble(coocc[3]));
				fouts[1].write(coocc[0] + "\t" + coocc[1] + "\t" + sig + NL);
				fouts[2].write(coocc[1] + "\t" + coocc[0] + "\t" + sig + NL);
				fouts[0].write(coocc[0] + "\t" + coocc[1] + "\t" + sig + NL);
				fouts[0].write(coocc[1] + "\t" + coocc[0] + "\t" + sig + NL); }
			fouts[0].close();
			fouts[1].close();
			fouts[2].close();
			fin.close();
			
			/* Run algorithm with specific feature data. */
			MySimRun sim12 = new MySimRun(fnbsymTmp.getPath(), fnb12Tmp.getPath(), fsim12.getPath());
			MySimRun sim21 = new MySimRun(fnbsymTmp.getPath(), fnb21Tmp.getPath(), fsim21.getPath());
				
			/* Create two threads. */
			Thread p12 = new Thread(sim12);
			Thread p21 = new Thread(sim21);
			
			/* Start both threads. */
			p12.start();
			p21.start();
				
			/* Check and update progress. */
			while (p12.isAlive() && p21.isAlive()) {
				setTaskState((sim12.getTaskProgress() + sim21.getTaskProgress()) / 2);
				Thread.sleep(2000); }
				
			/* Wait for termination. */
			p12.join();
			p21.join();
			printlnInfoEnd("... similarity calculated", tm);
			
			/* Combine both "edge" lists from similarity calculation.
			 * The following assumption is made:
			 *   - Similarity graphs are symmetric.
			 *   - Similarity graphs are sorted with respect to first word number (first column)
			 *     when considering only every second edge.
			 */
			setTaskState("Write graph 2", 10);
			printlnInfoBegin("Write graph 2 ...", tm);
			HashMap<IntegerPair, Integer> cooccs = new HashMap<IntegerPair, Integer>();
			
			BufferedReader fin12 = new BufferedReader(new FileReader(fsim12.getPath()));
			BufferedReader fin21 = new BufferedReader(new FileReader(fsim21.getPath()));
			BufferedWriter fout  = new BufferedWriter(new FileWriter(edgeFileName));
			
			int wtmp  = Nwmax + 1, wnmin = Nwmax, wnmax = 0;
			
			/* Read one line from second similarity graph. */ 
			String line21 = fin21.readLine();
			
			/* Read every second line from first similarity graph. */
			while ((line = fin12.readLine()) != null) {
				String[] coocc12 = line.split("\t");
				int w1 = Integer.parseInt(coocc12[0]);
				
				/* Next word in first similarity graph? */
				if (w1 != wtmp) {
					String[] coocc21 = line21.split("\t");
					int w3 = Integer.parseInt(coocc21[0]);
					
					/* Write all edges to cooccs with w3 >= w1 */
					while ((w3 >= wtmp) && (line21 != null)) {
						int w4     = Integer.parseInt(coocc21[1]);
						int weight = Integer.parseInt(coocc21[2]);
						
						/* Write edges which are present in both graphs. */
						if (weight >= Tnb) {
							IntegerPair p = new IntegerPair(w3, w4);
							Integer v;
							if ((v = cooccs.get(p)) != null) {
								weight += v;
								wnmin = Math.min(p.e1, Math.min(p.e2, wnmin));
								wnmax = Math.max(p.e1, Math.max(p.e2, wnmax));
								fout.write(p.e1 + "\t" + p.e2 + "\t" + weight + NL);
								fout.write(p.e2 + "\t" + p.e1 + "\t" + weight + NL); } }
							
						/* Skip one line and read the next one. */
						fin21.readLine();
						if ((line21 = fin21.readLine()) == null) {
							break; }
						coocc21 = line21.split("\t");
						w3 = Integer.parseInt(coocc21[0]); }
					
					/* Clear cooccs and continue with next word. */
					wtmp = w1;
					cooccs.clear(); }
				
				/* Continue with word w1 and save edge in cooccs. */
				else {
					int w2     = Integer.parseInt(coocc12[1]);
					int weight = Integer.parseInt(coocc12[2]);
					if (weight >= Tnb) { // before: (...) && (w1 <= w2)
						IntegerPair p = new IntegerPair(w1, w2);
						cooccs.put(p, weight); } }
				
				/* Skip one line. */
				fin12.readLine(); }
			fin12.close();
			fin21.close();
			fout.close();
			
			setTaskState(50);
			writeWordList(nodeFileName, Math.max(wnmin, Nbehead), wnmax);
			printlnInfoEnd("... files combined", tm);
			setTaskState(100); }
		catch (Exception e) {
			e.printStackTrace();
			throw new Exception("Cannot create similarity graph (graph 2): " + e.getMessage()); }
		return; }
	
	/**
	 * This method constructs the graph for the partitioning 2.
	 */
	@SuppressWarnings("unused")
	private void createGraph2_MySimBiem(String edgeFileName, String nodeFileName) throws Exception {
		Timer tm = new Timer();
		String hashFileName;
		BufferedReader fin;
		File fnb12Tmp  = File.createTempFile(fcorpus.getName() + ".nb12.graph", null);
		File fnb21Tmp  = File.createTempFile(fcorpus.getName() + ".nb21.graph", null);
		
		/* Add file names to the list of temporary files. */
		tmpFileList.add(fnb12Tmp.getAbsolutePath());
		tmpFileList.add(fnb21Tmp.getAbsolutePath());
		
		try {
			/*if (fsim12.exists() && fsim21.exists() && (doPart2.compareToIgnoreCase("force") != 0)) {
				printlnInfo("  Similarity files already exist (skip creation)."); }
			else {*/
			
			/* Use test graph .nbgraph.test if existent. */
			File fnbgraph = new File(fcorpus.getAbsolutePath() + ".nbgraph.test");
			if (fnbgraph.exists() && fnbgraph.isFile()) {
				hashFileName = fnbgraph.getAbsolutePath();
				printlnInfo("Use existing nb graph: " + fnbgraph.getPath() + "."); }
				
			/* Calculate cooccurrences. */
			else {
				setTaskState("Calculate cooccurrences", 0);
				printlnInfoBegin("Calculate cooccurrences ...", tm);
				MyCooccRun cAlg = new MyCooccRun();
				cAlg.start();
				cAlg = null;
				hashFileName = ConfigurationContainer.getPesistentHashFileName() + ".lgl2.expo";
				printlnInfoEnd("... cooccurrences calculated", tm); }
				
			/* Read calculated cooccurrences and construct graph 2. */
			setTaskState("Calculate similarity", 0);
			printlnInfoBegin("Calculate similarity ...", tm);
			BufferedWriter fout12 = new BufferedWriter(new FileWriter(fnb12Tmp));
			BufferedWriter fout21 = new BufferedWriter(new FileWriter(fnb21Tmp));
				
			/* Calculate similarity. */
			fin = new BufferedReader(new FileReader(hashFileName));
			String line;
			int neg = 0, pos = 0;
			while ((line = fin.readLine()) != null) {
				String[] coocc = line.split("\t");
				double dsig = Double.parseDouble(coocc[3]);
				if (dsig > 0.0) {
					pos++;
					int sig = (int) (100.0 * dsig);
					fout12.write(coocc[0] + "\t" + coocc[1] + "\t" + sig + NL);
					fout21.write(coocc[1] + "\t" + coocc[0] + "\t" + sig + NL); }
				else {
					neg++; } }
			//printlnInfo("Significance: " + pos + " (+), " + neg + " (-), " + (pos + neg) + " (total).");
			fout12.close();
			fout21.close();
			fin.close();
			
			/* Run algorithm with specific feature data. */
			MySimBiem sim = new MySimBiem(fnb12Tmp.getPath(), fnb21Tmp.getPath(), false, edgeFileName);
			
			/* Create two threads. */
			Thread p = new Thread(sim);
				
			/* Start thread. */
			p.start();
				
			/* Check and update progress. */
			while (p.isAlive()) {
				setTaskState(sim.getTaskProgress());
				Thread.sleep(1000); }
			printlnInfo("");
				
			/* Wait for termination. */
			p.join();
			printlnInfoEnd("... similarity calculated", tm);
			
			//writeWordList(nodeFileName, Math.max(wnmin, Nbehead), wnmax);
			writeWordList(nodeFileName, Nbehead, Nwmax);
			printlnInfoEnd("... files combined", tm);
			setTaskState(100); }
		catch (Exception e) {
			e.printStackTrace();
			throw new Exception("Cannot create similarity graph (graph 2): " + e.getMessage()); }
		return; }
	
	/**
	 * This method constructs the graph for the partitioning 2.
	 */
	@SuppressWarnings("unused")
	private void createGraph2_ok(String edgeFileName, String nodeFileName) throws Exception {
		Timer tm = new Timer();
		BufferedReader fin;
		File fnbsymTmp = File.createTempFile(fcorpus.getName() + ".nbsym.graph", null);
		File fnb12Tmp  = File.createTempFile(fcorpus.getName() + ".nb12.graph", null);
		File fnb21Tmp  = File.createTempFile(fcorpus.getName() + ".nb21.graph", null);
		
		tmpFileList.add(fnbsymTmp.getAbsolutePath());
		tmpFileList.add(fnb12Tmp.getAbsolutePath());
		tmpFileList.add(fnb21Tmp.getAbsolutePath());
		
		try {
			/*
			if (fsim12.exists() && fsim21.exists() && (doPart2.compareToIgnoreCase("force") != 0)) {
				printlnInfo("  Similarity files already exist (skip creation)."); }
			else {*/
			
			/* Use nb test graph .nbgraph.test if existent. */
			String hashFileName;
			File fnbgraph = new File(fcorpus.getAbsolutePath() + ".nbgraph.test");
			if (fnbgraph.exists() && fnbgraph.isFile()) {
				hashFileName = fnbgraph.getAbsolutePath();
				printlnInfo("Use existing nb graph: " + fnbgraph.getPath() + "."); }
				
			/* Calculate cooccurrences. */
			else {
				setTaskState("Calculate cooccurrences", 0);
				printlnInfoBegin("Calculate cooccurrences ...", tm);
				MyCooccRun cAlg = new MyCooccRun();
				cAlg.start();
				cAlg = null;
				printlnInfoEnd("... cooccurrences calculated", tm);
				hashFileName = ConfigurationContainer.getPesistentHashFileName() + ".lgl2.expo"; }
				
			/* Read calculated cooccurrences and construct graph 2. */
			setTaskState("Calculate similarity", 0);
			printlnInfoBegin("Calculate similarity ...", tm);
			BufferedWriter[] fouts = new BufferedWriter[3];
			fouts[0] = new BufferedWriter(new FileWriter(fnbsymTmp));
			fouts[1] = new BufferedWriter(new FileWriter(fnb12Tmp));
			fouts[2] = new BufferedWriter(new FileWriter(fnb21Tmp));
				
			fin = new BufferedReader(new FileReader(hashFileName));
			String line;
			int sig;
			while ((line = fin.readLine()) != null) {
				String[] coocc = line.split("\t");
				sig = (int) (100.0 * Double.parseDouble(coocc[3]));
				fouts[1].write(coocc[0] + "\t" + coocc[1] + "\t" + sig + NL);
				fouts[2].write(coocc[1] + "\t" + coocc[0] + "\t" + sig + NL);
				fouts[0].write(coocc[0] + "\t" + coocc[1] + "\t" + sig + NL);
				fouts[0].write(coocc[1] + "\t" + coocc[0] + "\t" + sig + NL); }
			fouts[0].close();
			fouts[1].close();
			fouts[2].close();
			fin.close();
			
			/* Run algorithm with specific feature data. */
			MySimRun sim12 = new MySimRun(fnbsymTmp.getPath(), fnb12Tmp.getPath(), fsim12.getPath());
			MySimRun sim21 = new MySimRun(fnbsymTmp.getPath(), fnb21Tmp.getPath(), fsim21.getPath());
				
			/* Create two threads. */
			Thread p12 = new Thread(sim12);
			Thread p21 = new Thread(sim21);
			
			/* Start both threads. */
			p12.start();
			p21.start();
				
			/* Check and update progress. */
			while (p12.isAlive() && p21.isAlive()) {
				setTaskState((sim12.getTaskProgress() + sim21.getTaskProgress()) / 2);
				Thread.sleep(2000); }
				
			/* Wait for termination. */
			p12.join();
			p21.join();
			printlnInfoEnd("... similarity calculated", tm);
			
			/* Combine both "edge" lists from similarity calculation. */
			setTaskState("Write graph 2", 10);
			printlnInfoBegin("Write graph 2 ...", tm);
			HashMap<IntegerPair, Integer> cooccs = new HashMap<IntegerPair, Integer>();
			fin = new BufferedReader(new FileReader(fsim12.getPath()));
			
			while ((line = fin.readLine()) != null) {
				String[] coocc = line.split("\t");
				int w1 = Integer.parseInt(coocc[0]);
				int w2 = Integer.parseInt(coocc[1]);
				int weight = Integer.parseInt(coocc[2]);
				if ((weight >= Tnb) && (w1 <= w2)) {
					IntegerPair p = new IntegerPair(w1, w2);
					cooccs.put(p, weight); } }
			fin.close();
			setTaskState(50);
			
			/* Write edges which are present in both graphs. */
			fin = new BufferedReader(new FileReader(fsim21.getPath()));
			BufferedWriter fout = new BufferedWriter(new FileWriter(edgeFileName));
			int wnmin = Nwmax, wnmax = 0;
			Integer v;
			while ((line = fin.readLine()) != null) {
				String[] coocc = line.split("\t");
				int w1 = Integer.parseInt(coocc[0]);
				int w2 = Integer.parseInt(coocc[1]);
				
				int weight = Integer.parseInt(coocc[2]);
				if ((weight >= Tnb) && (w1 <= w2)) {
					IntegerPair p = new IntegerPair(w1, w2);
					if ((v = cooccs.get(p)) != null) {
						weight += v.intValue();
						wnmin = Math.min(w1, Math.min(w2, wnmin));
						wnmax = Math.max(w1, Math.max(w2, wnmax));
						fout.write(w1 + "\t" + w2 + "\t" + weight + NL);
						fout.write(w2 + "\t" + w1 + "\t" + weight + NL); } } }
			fin.close();
			fout.close();
			writeWordList(nodeFileName, Math.max(wnmin, Nbehead), wnmax);
			printlnInfoEnd("... files combined", tm);
			setTaskState(100); }
		catch (Exception e) {
			e.printStackTrace();
			throw new Exception("Cannot create similarity graph (graph 2): " + e.getMessage()); }
		return; }
	
	/**
	 * This method creates a graph where nodes are clusters from partitioning 1 and 2, and edge weights
	 * are the number of common members (words) in the specific nodes (clusters).
	 *
	 * <strong>Note:</strong> The node list in graph 2 starts with word number Nbehead. In the corresponding
	 *                        cluster file word numbers start with 1. In this method word numbers from the
	 *                        cluster file are adjusted by plus Nbehead.
	 *
	 * @param fp1name      the name of the file with clusters from partitioning 1.
	 * @param fp2name      the name of the file with clusters from partitioning 2.
	 * @param edgeFileName the output file name of the edge list.
	 * @param nodeFileName the output file name of the node list.
	 */
	private void createGraph3(String fp1name, String fp2name, String edgeFileName, String nodeFileName) {
		
		setTaskState("Get overlapping clusters", 0);
		try {
			/* Words and A-cluster numbers (partitioning 1). */
			HashMap<Integer, Integer> wxcA = new HashMap<Integer, Integer>();
			String line;
			String[] word;
			
			/* Read first A-cluster file (partitioning 1): word[0] word number, word[2] cluster number. */
			BufferedReader fin = new BufferedReader(new FileReader(fp1name));
			while ((line = fin.readLine()) != null) {
				word = line.split("\t");
				wxcA.put(Integer.parseInt(word[0]), Integer.parseInt(word[2])); }
			fin.close();
			setTaskState(25);
			
			/* Read second cluster file (partitioning 2, B-cluster). */
			fin = new BufferedReader(new FileReader(fp2name));
			HashMap<IntegerPair, Integer> overlap  = new HashMap<IntegerPair, Integer>();
			Integer wn, cnA, cnB, noverlap;
			while ((line = fin.readLine()) != null) {
				word = line.split("\t");
				wn   = Integer.parseInt(word[0]) + Nbehead - 1; /* Word number adjustement. !!!test!!! */
				cnB  = Integer.parseInt(word[2]);
				
				/* Count overlap if word wn is already member of an A-cluster. */
				if ((cnA = wxcA.get(wn)) != null) {
					IntegerPair cp = new IntegerPair(cnA.intValue(), cnB.intValue());
					
					/* Count number of matches in cluster A and B. */
					if ((noverlap = overlap.get(cp)) == null) {
						overlap.put(cp, 1); }
					else {
						overlap.put(cp, noverlap.intValue() + 1); } } }
			fin.close();
			wxcA.clear(); wxcA = null;
			setTaskState(70);
			
			/* Now write clusters as nodes to file. */
			BufferedWriter fnodes = new BufferedWriter(new FileWriter(nodeFileName));
			BufferedWriter fedges = new BufferedWriter(new FileWriter(edgeFileName));
			HashMap<String, Integer> clusterlist = new HashMap<String, Integer>();
			int nnode = 1; /* new node IDs. */
			for (IntegerPair cp : overlap.keySet()) {
				
				/* If cluster A and B do overlap. */
				if (((noverlap = overlap.get(cp)) != null) && (noverlap.intValue() >= Noverlap)) {
					
					String cA = "A" + cp.e1;
					String cB = "B" + cp.e2;
					
					/* Check if cluster A and B are already registered. */
					if (clusterlist.get(cA) == null) {
						fnodes.write(nnode + "\t" + cA + NL);
						clusterlist.put(cA, nnode++); }
					if (clusterlist.get(cB) == null) {
						fnodes.write(nnode + "\t" + cB + NL);
						clusterlist.put(cB, nnode++); }
					
					int ncA = clusterlist.get(cA).intValue();
					int ncB = clusterlist.get(cB).intValue();
					
					fedges.write(ncA + "\t" + ncB + "\t" + noverlap.intValue() + NL);
					fedges.write(ncB + "\t" + ncA + "\t" + noverlap.intValue() + NL); } }
			
			if (nnode <= 1) {
				stopHere("No overlap between partitioning 1 and 2 (no edges in graph 3)."); }
			
			fnodes.close();
			fedges.close();
			setTaskState(100); }
		catch (Exception e) {
			printlnError("Exception: " + e.getMessage()); }
		return; }
	
	/**
	 * 
	 * @return <code>true</code> if the preprocessing is required, otherwise <code>false</code>.
	 */
	private boolean needPreprocessing() {
		if (doPreproc.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (preprocessing).");
			return false; }
		if (doPreproc.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip preprocessing.");
			return false; }
		return true; }
	
	/**
	 * This method determines if the calculation of partitioning 1 is necessary or not. The result only
	 * depends on the configuration, not on the existence of graph-1 or cluster-1 files. For more information
	 * see the manual UnsuPos.pdf or cf. option <code>PART1</code> in the configuration file (options are
	 * <code>true</code>, <code>false</code>, <code>force</code>, <code>break</code>). 
	 * 
	 * @return <code>true</code> if the calculation of partitioning 1 is necessary, otherwise
	 *         <code>false</code>.
	 */
	private boolean needPartitioning1() {
		if (doPart1.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (partitioning 1).");
			return false; }
		if (doPart1.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip partitioning 1.");
			return false; }
		return true; }
	
	/**
	 * This method determines if the calculation of partitioning 2 is necessary or not. The result only
	 * depends on the configuration, not on the existence of graph-2 or cluster-2 files.
	 * 
	 * @return <code>true</code> if the calculation of partitioning 2 is necessary, otherwise
	 *         <code>false</code>.
	 * 
	 * @see needPartitioning1()
	 */
	private boolean needPartitioning2() {
		if (doPart2.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (partitioning 2).");
			return false; }
		if (doPart2.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip partitioning 2.");
			return false; }
		return true; }
	
	private boolean needJoinPartitioning() {
		if (doJoinParts.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (joining partitionings).");
			return false; }
		if (doJoinParts.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip joining of partitionings.");
			return false; }
		return true; }
	
	private boolean needBuildLexicon() {
		if (doBuildLex.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (lexicon building).");
			return false; }
		if (doBuildLex.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip building of lexicon.");
			return false; }
		return true; }
	
	private boolean needTagging() {
		if (doTagg.compareToIgnoreCase("break") == 0) {
			stopHere("Stop execution (Viterbi tagging).");
			return false; }
		if (doTagg.compareToIgnoreCase("false") == 0) {
			printlnInfo("Skip tagging.");
			return false; }
		return true; }
	
	/**
	 * This method indicates if the calculation of the cosine similarity graph is necessary or not.
	 * 
	 * @return <code>true</code> when the calculation of the cosine similarity graph is necessary,
	 *         otherwise <code>false</code>.
	 */
	private boolean needCosGraph() {
		if (fcosgraph.exists() && (doPart1.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Sorted cosine similarity graph already exists (skip creation).");
			return false; }
		return true; }
	
	/**
	 * This method indicates if the calculation of graph 1 is necessary or not.
	 * 
	 * @return <code>true</code> if the calculation of graph 1 is necessary, otherwise <code>false</code>.
	 */
	private boolean needGraph1() {
		if (new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.nodes").exists() &&
			new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.edges").exists() &&
			(doPart1.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Graph 1 already exists (skip creation).");
			return false; }
		return true; }
	
	/**
	 * This method indicates if the calculation of graph 2 is necessary or not.
	 * 
	 * @return <code>true</code> if the calculation of graph 2 is necessary, otherwise <code>false</code>.
	 */
	private boolean needGraph2() {
		if (new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.nodes").exists() &&
			new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.edges").exists() &&
			(doPart2.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Graph 2 already exists (skip creation).");
			return false; }
		return true; }
	
	/**
	 * This method checks whether (re-)clustering of graph 1 is required or not. Graph 1 will be clustered
	 * when cluster files do not exist or when a recalculation of partitioning 1 is forced.
	 * 
	 * @return <code>true</code> when graph 1 has to be (re-)clustered, otherwise <code>false</code>.
	 */
	private boolean needClustering1() {
		if (new File(dirGraph1, fcorpus.getName() + ".graph1.cluster").exists()      &&
			new File(dirGraph1, fcorpus.getPath() + ".graph1.cluster.read").exists() &&
			(doPart1.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Cluster files already exist (skip CW clustering of graph 1).");
			return false; }
		return true; }
	
	/**
	 * This method checks whether (re-)clustering of graph 2 is required or not. Graph 2 will be clustered
	 * when cluster files do not exist or when a recalculation of partitioning 2 is forced.
	 * 
	 * @return <code>true</code> when graph 2 has to be (re-)clustered, otherwise <code>false</code>.
	 */
	private boolean needClustering2() {
		if (new File(dirGraph2, fcorpus.getName() + ".graph2.cluster").exists()      &&
			new File(dirGraph2, fcorpus.getPath() + ".graph2.cluster.read").exists() &&
			(doPart2.compareToIgnoreCase("force") != 0)) {
			printlnInfo("  Cluster files already exist. Skip CW clustering.");
			return false; }
		return true; }
	
	/**
	 * This method deletes temporary files.
	 */
	private void cleanUp() {
		if (deleteTempFiles) {
			printlnInfo("Delete temporary files ...");
			
			/* Delete temp. word number and frequency files. */
			String[] suffixes = {".bow", ".bin", ".idx", ".meta", ".tmp", ".sort_wordStr", ".trie",
					             ".renumbered", ".renumbered.bin", ".renumbered.idx",
					             ".renumbered.meta", ".renumbered.tmp" };
			
			for (int i = 0; i < tmpFileList.size(); i++) {
				deleteFileIfExists(tmpFileList.get(i));
				for (String suffix : suffixes) {
					deleteFileIfExists(tmpFileList.get(i) + suffix); } }
			
			deleteFileIfExists(fcorpus.getPath() + ".tmodel.lexicon.front.pretree");
			deleteFileIfExists(fcorpus.getPath() + ".tmodel.lexicon.rev.pretree");
			deleteFileIfExists(fcorpus.getPath() +
				".de.uni_leipzig.asv.medusa.loader.filter.IDXNeighbourhoodFilterImpl.hash");
			deleteFileIfExists(fcorpus.getPath() +
		        ".de.uni_leipzig.asv.medusa.loader.filter.IDXNeighbourhoodFilterImpl.hash.lgl.expo");
			deleteFileIfExists(fcorpus.getPath() +
			    ".de.uni_leipzig.asv.medusa.loader.filter.IDXNeighbourhoodFilterImpl.hash.lgl2.expo"); }
		return;	}
	
	/**
	 * This method deletes all data files.
	 */
	public void deleteDataFiles() {
		String dir;
	
		/* Delete all from the basic directory. */
		dir = dirBasic.getPath();
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".tok").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".wn").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".wf").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".internal.wn").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".internal.wf").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".wcos_sorted_(2_0_1).edges").getAbsolutePath());
		
		/* Delete all from graph1/ */
		dir = dirGraph1.getPath();
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph1.nodes").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph1.edges").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph1.cluster").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph1.cluster.read").getAbsolutePath());
		
		/* Delete all from graph2/ */
		dir = dirGraph2.getPath();
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph2.nodes").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph2.edges").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph2.cluster").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph2.cluster.read").getAbsolutePath());
		
		/* Delete all from graph3/ */
		dir = dirGraph3.getPath();
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph3.nodes").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph3.edges").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph3.cluster").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".graph3.cluster.read").getAbsolutePath());
		
		dir = dirTmodel.getPath();
		deleteFileIfExists(fcr.getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".tmodel.model").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".tmodel.taglist").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".tmodel.lexicon").getAbsolutePath());
		deleteFileIfExists(new File(dir, fcorpus.getName() + ".tmodel.transitions").getAbsolutePath());
		deleteFileIfExists(new File(fcorpus.getName() + ".fc").getAbsolutePath());
		return; }
	
	/**
	 * This method delets a file from the file system.
	 * 
	 * @param fpath the name of the file to be deleted.
	 */
	private void deleteFileIfExists(String fpath) {
		File f = new File(fpath);
		if (f.exists() && f.isFile()) {
			if (f.delete()) {
				printlnInfo("Delete file " + f.getPath() + "."); }
			else {
				printlnError("File " + f.getPath() + " can not be deleted."); } }
		return; }
	
	/**
	 * This method creates a words list with word numbers and frequency, respectively. The output is
	 * stored in files with suffix "wn" and "wf", respectively.<p>
	 * For the creation of the word and frequency list the <code>WordNumbers</code> class from the
	 * Medusa project is used.<p>
     *
     * <strong>See also</strong>: de.uni_leipzig.asv.medusa<p>
	 *                            de.uni_leipzig.asv.medusa.input.WordNumbers
	 */
	private void createFrequencyList() throws Exception {
		
		/* if (!fcorpus.exists() || !fcorpus.isFile()) 
		 * if (ftok.exists() && (doPreproc.compareToIgnoreCase("force") != 0)) */
		setTaskState("Preprocessing", "Create word and frequency list", 5);
		
		File fwnumb = new File(dirBasic.getPath(), fcorpus.getName() + ".wn");
		File fwfreq = new File(dirBasic.getPath(), fcorpus.getName() + ".wf");
		
		/* Copy existing word and frequency list to temp directory. */
		/*if (fwnumb.exists() && fwfreq.exists() && (doPreproc.compareToIgnoreCase("force") != 0)) {
			printlnInfo("Word and frequency list already exist (skip creation).");
			setTaskState(20);
			copyFile(fwnumb.getPath(), fwn.getPath());
			setTaskState(55);
			copyFile(fwfreq.getPath(), fwf.getPath()); }*/
		
		/* Create a list of words with word number and frequency, respectively. */
		try {
			
		    /* Always create an up to date word/frequency list. This is necessary since the
		     * tokenized file can be created by the user, not by the <code>UnsuPosTagger</code>
		     * which means that word number and frequency list must be recreated. */
			printInfoAlienOutputBegin("Medusa output");
			WordNumbers wn = new WordNumbers();
			wn.init();
			setTaskState(10);
			wn.startReading();
				
			/* Transform internal to external word numbers (Medusa!). */
			setTaskState(20);
			setGlobalProperty("MEMORY_ALLOCATOR_IMPL",
	                          "de.uni_leipzig.asv.medusa.config.DefaultMemoryAllocatorImpl");
			Transform trans = new WordNumberTransformerImpl();
			trans.init();
			trans.transform();
				
			printInfoAlienOutputEnd();
			setTaskState(80);
			
			/* Write a copy of word and frequency list to the basic directory. */
			copyFile(fwn.getPath(), fwnumb.getPath()); setTaskState(90);
			copyFile(fwf.getPath(), fwfreq.getPath()); }
		catch (ConfigurationException e) {
			throw new Exception(e); }
		catch (Exception e) {
			throw new Exception(e); }
		setTaskState(100);
	    return; }
	
	/**
	 * This method starts the Chinese Whispers clustering algorithm for the specified graph.
	 * The CW algorithm works with a copy of the node and edge file in the temporary directory.
	 * 
	 * @param fnodes the file name of the node list.
	 * @param fedges the file name of the edge list.
	 * @param fout   the output file name.
	 * 
	 * @throws Exception
	 */
	private void runCWClustering(String fnodes, String fedges, String fout) throws Exception {
		try {
			setTaskState("CW clustering", 0);
			File tmpFileNodes = File.createTempFile(new File(fnodes).getName(), null);
			File tmpFileEdges = File.createTempFile(new File(fedges).getName(), null);
			
			/* Add file names to the list of temporary files. */
			tmpFileList.add(tmpFileNodes.getAbsolutePath());
			tmpFileList.add(tmpFileEdges.getAbsolutePath());
			
			/* Copy node and edge file to the temporary directory. */
			copyFile(fnodes, tmpFileNodes.getPath());
			copyFile(fedges, tmpFileEdges.getPath());
			setTaskState(20);
			
			printInfoAlienOutputBegin("CW output");
			String[] args = {"-F", "-i", tmpFileNodes.getPath(), tmpFileEdges.getPath(),
					         "-a", "top", "-r", "-o", fout};
			de.uni_leipzig.asv.toolbox.ChineseWhispers.main.Start.main(args);
			printInfoAlienOutputEnd();
			setTaskState(100); }
		catch (Exception e) {
			throw new Exception(e); }
		return; }
	
	/**
	 * This method resolves the clusters of cluster nodes into clusters of words.
	 * 
	 * @param fnameA
	 * @param fnameB
	 * @param fnameAB
	 */
	private void resolveCluster(String fnameA, String fnameB, String fnameAB) {
		
		setTaskState("Resolve clusters", 0);
		try {
			/* Read list of A-clusters (from partitioning 1). */
			BufferedReader fin = new BufferedReader(new FileReader(fnameA));
			HashMap<Integer, String> clusterA = new HashMap<Integer, String>();
			String line;
			String[] cluster;
			while ((line = fin.readLine()) != null) {
				cluster = line.split("\t");
				clusterA.put(Integer.parseInt(cluster[0]), cluster[2]); }
			fin.close();
			setTaskState(25);
			
			/* Read list of B-clusters (from partitioning 2). */
			fin = new BufferedReader(new FileReader(fnameB));
			HashMap<Integer, String> clusterB = new HashMap<Integer, String>();
			while ((line = fin.readLine()) != null) {
				cluster = line.split("\t");
				clusterB.put(Integer.parseInt(cluster[0]), cluster[2]); }
			fin.close();
			setTaskState(50);
			
			/* Resolve clusters A.../B... into words. */
			fin = new BufferedReader(new FileReader(fnameAB));
			BufferedWriter fout = new BufferedWriter(new FileWriter(fcr));
			int maxcno = -1;
			HashMap<String, Integer> topwords = readWordList(1, Nsingadd);
			String[] words;
			while ((line = fin.readLine()) != null) {
				cluster = line.split("\t");
				maxcno = Math.max(maxcno, Integer.parseInt(cluster[0]));
				
				/* Resolve the clusters. */
				//ArrayList<String> clusterMembers = new ArrayList<String>();
				HashSet<String> clusterMembers = new HashSet<String>();
				for (String c : cluster[2].split(", ")) {
					int cno = Integer.parseInt(c.substring(1, c.length()));
					
					/* Get words from A-clusters, or ... */
					if (c.substring(0, 1).equals("A")) {
						words = clusterA.get(cno).split(", ");
						clusterA.remove(cno); }
					
					/* ... get words from B-clusters. */
					else {
						words = clusterB.get(cno).split(", ");
						clusterB.remove(cno); }
					
					for (String word : words) {
						clusterMembers.add(word); } }
				
				/* Now write all cluster Members. */
				fout.write(cluster[0] + "\t" + clusterMembers.size() + "\t");
				String sep = "";
				Iterator<String> it = clusterMembers.iterator();
				while (it.hasNext()) {
					String word = (String) it.next();
					topwords.remove(word);
					fout.write(sep + word);
					sep = ", "; }
				
				clusterMembers.clear(); clusterMembers = null;
				fout.write(NL); }
			fin.close();
			setTaskState(80);
			
			/* Write all remaining A-clusters (i.e. from first partitioning). */
			for (int cno : clusterA.keySet()) {
				maxcno++;
				words = clusterA.get(cno).split(", ");
				for (String word : words) {
					topwords.remove(word); }
				fout.write(maxcno + "\t" + words.length + "\t" + clusterA.get(cno) + NL); }
			setTaskState(95);
			
			/* Write all remaining top words. */
			for (String word : topwords.keySet()) {
				maxcno++;
				fout.write(maxcno + "\t1\t" + word + NL); }
			setTaskState(100);
			fout.close(); }
		catch (Exception e) {
			printlnError("Exception: " + e.getMessage()); }
		return; }
	
	/**
	 * This methods reads words and corresponding word numbers from file.
	 *  
	 * @param  fname the name of the word list file
	 * @param  N1 the lowest word number
	 * @param  N2 the highest word number
	 * @return a map of words to word numbers
	 * @throws Exception
	 */
	private HashMap<String, Integer> readWordList(int N1, int N2) throws Exception {
		int hashSize = (int) (1.33 * Math.abs(N2 - N1));
		HashMap<String, Integer> wordlist = new HashMap<String, Integer>(hashSize);
		String word;
		
		try {
			BufferedReader fin = new BufferedReader(new FileReader(fwn.getPath()));
			int n = 0;
		    while ((word = fin.readLine()) != null) {
		    	n++;
		    	if (n > N2) { break;    }
		    	if (n < N1) { continue; }
	    		String[] w = word.split("\t");
	    		if (w.length != 2) {
	    			throw new Exception("Format error in wordlist " + fwn.getPath() +
	    					            " at line " + n + "."); }
	    		wordlist.put(w[1], new Integer(w[0])); } }
		catch (FileNotFoundException e) {
			throw new FileNotFoundException("Word list file not found (" + fwn.getPath() + ")."); }
		
		return wordlist; }
	
	/**
	 * This method reads the frequency list from file into a HashMap.
	 * 
	 * @param N1 the lowest word number from which file will be read.
	 * @param N2 the highest word number up to which file is read.
	 * @return a <code>HashMap<String, Integer></code> which maps words to frequency.
	 * @throws Exception
	 */
	private HashMap<String, Integer> readFreqList(int N1, int N2) throws Exception {
		
		/* Test index range. */
		if (N2 <= N1) {
			throw new Exception(); }
		
		/* Calculate hash size (assuming load factor of approx. 0.75). */
		int N = (int) (1.33 * (double) (N2 - N1 + 1));
		
		HashMap<String, Integer> freqList = new HashMap<String, Integer>(N);
		String line1, line2;
		try {
			BufferedReader fwNumber = new BufferedReader(new FileReader(fwn.getPath()));
			BufferedReader fwFreq   = new BufferedReader(new FileReader(fwf.getPath()));
			int n = 0;
			while (((line1 = fwNumber.readLine()) != null) && ((line2 = fwFreq.readLine()) != null)) {
				n++;
				if (n < N1) { continue; }
				if (n > N2) { break;    }
				String[] wn = line1.split("\t");
				String[] wf = line2.split("\t");
				if (!wn[0].equals(wf[0])) {
					throw new Exception("Word and frequency list do not match at line " + n + "."); }
				freqList.put(wn[1], new Integer(wf[1])); }
			fwNumber.close();
			fwFreq.close(); }
		catch (FileNotFoundException e) {
			printlnError("Exception: " + e.getMessage());
			throw new FileNotFoundException("Exception: " + e.getMessage()); }
		return freqList; }
	
	private void writeWordList(String outfilename, int N1, int N2) {
		try {
			BufferedReader fin  = new BufferedReader(new FileReader(fwn.getPath()));
			BufferedWriter fout = new BufferedWriter(new FileWriter(outfilename));
			String word;
			int n = 0;
			while ((word = fin.readLine()) != null) {
				n++;
				if (n >= N1) { fout.write(word + NL); }
				if (n >  N2) { break; } }
			fin.close();
			fout.close(); }
		catch (FileNotFoundException e) {
			printlnError("Exception: " + e.getMessage()); }
		catch (IOException e) {
			printlnError("Exception: " + e.getMessage()); }
		return; }
	
	/**
	 * This method reads a frequency list (word number, frequency) from the file specified by the parameter
	 * <code>fname</code>.
	 * 
	 * @param  fname the file name of the frequency list.
	 * @return A hashmap of word numbers and corresponding frequency.
	 * @throws Exception
	 */
	@SuppressWarnings("unused")
	private HashMap<Integer, Integer> readWordNumberFreqList(String fname) throws Exception {
		HashMap<Integer, Integer> freqlist = new HashMap<Integer, Integer>(); /* !!! */
		String word;
		
		try {
			BufferedReader fin = new BufferedReader(new FileReader(fname));
			int n = 0;
		    while ((word = fin.readLine()) != null) {
		    	n++;
		    	String[] w = word.split("\t");
		    	if (w.length != 2) {
		    		printlnError("Format error in frequency list "  + fname + " at line " + n);
		    		throw new Exception("Format error in frequency list " + fname + " at line " + n); }
		    	freqlist.put(new Integer(w[0]), new Integer(w[1])); } }
		catch (FileNotFoundException e) {
			printlnError("Frequency list not found: " + fname);
			throw new FileNotFoundException("Frequency list not found: " + fname); }
		
		return freqlist; }
	
	/**
	 * This method writer the first N+ edges from file <code>finame</code> to file <code>foname</code>.<p>
	 * 
	 * <strong>Note:</strong> The third column (weight values) of the input file must be sorted (desc). 
	 * 
	 * @param finame     the name of the input file (cosine similarity graph).
	 * @param foname     the name of the output file (same format as input file).
	 * @param N          the number of nodes to write (the resulting number of nodes can be greater, since
	 *                   nodes connected by edges with weights equal to the smallest weight in the graph
	 *                   are included as well).
	 * @throws Exception
	 */
	private int writeFirstNodes(String finame, String foname, int N) throws Exception {
		Timer tm = new Timer();
		String line;
		int maxNodeId = -1;
		
		try {
			printlnInfoBegin("Determine first " + N + "+ nodes ...", tm);
			BufferedReader fin  = new BufferedReader(new FileReader(finame));
			BufferedWriter fout = new BufferedWriter(new FileWriter(foname));
		
			/* Init the node list. */
			HashMap<String, Boolean> nodelist = new HashMap<String, Boolean>();
			int n = 0; int Nnodes = 0;
			int lastWeight = -1;
			while ((line = fin.readLine()) != null) {
				n++;
				String[] edge = line.split("\t");
				if (edge.length != 3) {
					printlnError("Format error in edge list " + fcorpus.getPath() + ".edges" +
								 " at line " + n);
					throw new Exception("Format error in wordlist " + fcorpus.getPath() + ".graph1.edges" +
			                            " at line " + n); }
				if (Nnodes <= N) {
					if (nodelist.get(edge[0]) == null) {
						Nnodes++;
						nodelist.put(edge[0], true); }
					if (nodelist.get(edge[1]) == null) {
						Nnodes++;
						nodelist.put(edge[1], true); } }
				int weight = Integer.parseInt(edge[2]);
				if ((Nnodes > N) && (weight < lastWeight)) {
					break; }
				maxNodeId = Math.max(maxNodeId, Integer.parseInt(edge[0]));
				maxNodeId = Math.max(maxNodeId, Integer.parseInt(edge[1]));
				fout.write(edge[0] + "\t" + edge[1] + "\t" + weight + NL);
				lastWeight = weight; }
			printlnInfo("  " + n + " lines read from file.");
			fout.close();
			
			/* Read number of nodes in graph. */
			printlnInfo("  " + nodelist.size() + " nodes in graph.");
			nodelist.clear();
			printlnInfoEnd("... done", tm); }
		catch (Exception e) {
			throw new Exception(e); }
		return maxNodeId; }
	
	/**
	 * This method writer the first N+ edges from file <code>finame</code> to file <code>foname</code>.<p>
	 * 
	 * <strong>Note:</strong> The third column (weight values) of the input file must be sorted (desc). 
	 * 
	 * @param finame
	 * @param foname
	 * @param N
	 * @throws Exception
	 */
	@SuppressWarnings("unused")
	private int writeFirstNodes_prev(String finame, String foname, int N) throws Exception {
		Timer tm = new Timer();
		String line;
		int maxNodeId = -1;
		
		try {
			printlnInfoBegin("Determine first " + N + "+ nodes ...", tm);
			BufferedReader fin  = new BufferedReader(new FileReader(finame));
			BufferedWriter fout = new BufferedWriter(new FileWriter(foname));
		
			/* Init the node list. */
			boolean[] nodelist = new boolean[N + 1];
			for (int i = 0; i <= N; i++) { nodelist[i] = false; }
			int n = 0; int Nnodes = 0;
			int lastWeight = -1;
			while ((line = fin.readLine()) != null) {
				n++;
				String[] edge = line.split("\t");
				if (edge.length != 3) {
					printlnError("Format error in edge list " + fcorpus.getPath() + ".edges" +
								 " at line " + n);
					throw new Exception("Format error in wordlist " + fcorpus.getPath() + ".graph1.edges" +
			                            " at line " + n); }
				int node1  = Integer.parseInt(edge[0]);
				int node2  = Integer.parseInt(edge[1]);
				
				if (Nnodes <= NCWtarg) {
					if (nodelist[node1] == false) { Nnodes++; nodelist[node1] = true; }
					if (nodelist[node2] == false) { Nnodes++; nodelist[node2] = true; } }
				int weight = Integer.parseInt(edge[2]);
				if ((Nnodes > NCWtarg) && (weight < lastWeight)) {
					break; }
				maxNodeId = Math.max(maxNodeId, Math.min(node1, node2));
				fout.write(node1 + "\t" + node2 + "\t" + weight + NL);
				lastWeight = weight; }
			printlnInfo("  " + n + " lines read from file.");
			fout.close();
			
			/* Read number of nodes in graph. */
			n = 0;
			for (int i = 0; i <= N; i++) {
				if (nodelist[i]) { n++; } }
			printlnInfo("  " + n + " nodes in graph.");
			printlnInfoEnd("... done", tm); }
		catch (Exception e) {
			throw new Exception(e); }
		return maxNodeId; }
	
	/**
	 * This method determines the maximal word number in the cluster files from partitioning 1 and 2,
	 * respectively.
	 * 
	 * <strong>Note:</strong> The node list in graph 2 starts with word number Nbehead. In the corresponding
	 *                        cluster file word numbers start with 1. In this method word numbers from the
	 *                        cluster file are adjusted by plus Nbehead.
	 * 
	 * @return the biggest word number in the cluster files.
	 * @throws Exception
	 */
	private int getMaxWordNumberFromClusterFiles() throws Exception {
		String line;
		int nmax = -1;
		try {
			File f = new File(dirGraph1.getPath(), fcorpus.getName() + ".graph1.cluster");
			BufferedReader fin = new BufferedReader(new FileReader(f.getPath()));
			while ((line = fin.readLine()) != null) {
				String[] word = line.split("\t");
				nmax = Math.max(nmax, Integer.parseInt(word[0])); }
			fin.close();
			f   = new File(dirGraph2.getPath(), fcorpus.getName() + ".graph2.cluster");
			fin = new BufferedReader(new FileReader(f.getPath()));
			while ((line = fin.readLine()) != null) {
				String[] word = line.split("\t");
				nmax = Math.max(nmax, Integer.parseInt(word[0]) + Nbehead); } }
		catch (Exception e) {
			throw new Exception("Cannot determine the maximal word number in the cluster files."); }
		return nmax; }
	
	/**
	 * This method determines the highest tag (cluster) number of the joined partitioning in file
	 * <code>*.cluster.read</code>.
	 * 
	 * @return the highest tag (cluster) number from <code>*.cluster.read</code>. 
	 */
	private int getMaxTagNumber() throws Exception {
		String line;
		int N = -1;
		
		try {
			BufferedReader fin = new BufferedReader(new FileReader(fcr));
			while ((line = fin.readLine()) != null) {
				String[] cluster = line.split("\t");
				N = Math.max(N, Integer.parseInt(cluster[0])); }
			fin.close(); }
		catch (FileNotFoundException e) {
			throw new Exception("Exception: " + e.getMessage()); }
			
		/* Check the result. */
		if (N <= 0) {
			throw new Exception("Exception: Cannot read tags from file " + fcr.getPath()); }
		return N; }
	
	private void printlnInfoBegin(String s, Timer tm) {
		System.out.println(System.getProperty("line.separator") + s);
		if (sout != null) {
			sout.println(s); }
		if (tm != null) { tm.start(); }
		return; }
	
	private void printlnInfoEnd(String s, Timer tm) {
		String msg = s;
		if (tm != null) {
			tm.stop();
			msg = s + " (" + tm.toString() + ")."; }
		System.out.println(msg);
		if (sout != null) {
			sout.println(msg); }
		return; }
	
	private void printInfo(String s) {
		System.out.print(s);
		System.out.flush();
		if (sout != null) {
			sout.print(s);
			sout.flush(); }
		return;	}
	
	private void printlnInfo(String s) {
		System.out.println(s);
		if (sout != null) {
			sout.println(s); }
		return;	}
	
	private void printlnError(String s) {
		System.err.println(s);
		if (eout != null) {
			eout.println(s); }
		return; }
	
	/**
	 * This method prints a line on standard output indicating the begin of output from other package's
	 * methods.
	 * 
	 * @param s a string which is printed to describe the content of the subsequent output
	 */
	private void printInfoAlienOutputBegin(String s) {
		String msg = "...";
		strInfoAlienOutput = s;
		if (s.length() > 0) { msg = "Begin of " + s; }
		System.out.println("========== " + msg + " =========="); }
	
	/**
	 * This method prints a line on standard output indicating the end of output from other package's
	 * methods.
	 */
	private void printInfoAlienOutputEnd() {
		String msg = "...";
		if (strInfoAlienOutput.length() > 0) { msg = "End of " + strInfoAlienOutput; }
		strInfoAlienOutput = "";
		System.out.println("------------ " + msg + " ----------"); }
	
	/**
     * This method sets the name of the actual step.
     * 
     * @param name a description for the actual step performed.
     */
    @SuppressWarnings("unused")
	private void setStepName(String name) {
    	stepName     = name;
    	taskName     = "";
    	taskProgress = 0;
    	return; }
    
    /**
	 * This method sets the informal state about the progress of the actual task (step and task name, and
	 * task progress). One task normally consists of several tasks.
	 * 
	 * @param stepName a description of the actual step.
	 * @param taskName a description of the actual task.
	 * @param i        a number describing the progress of the actual task (between 0 and 100).
	 */
	private void setTaskState(String stepName, String taskName, int i) {
		
		/* Set task name and task state. */
		this.stepName     = stepName;
		this.taskName     = taskName;
		this.taskProgress = Math.max(Math.min(i, 100), 0);
		return; }
    
    /**
	 * This method sets the informal state about the progress of the actual task (task name and task
	 * progress).
	 * 
	 * @param name a description of the actual task.
	 * @param i    a number describing the progress of the actual task (between 0 and 100).
	 */
	private void setTaskState(String name, int i) {
		
		/* Set task name and task state. */
		taskName     = name;
		taskProgress = Math.max(Math.min(i, 100), 0);
		return; }
    
    /**
	 * This method sets the informal state about the progress of the actual task progress.
	 * 
	 * @param i a number describing the progress of the actual task (between 0 and 100).
	 */
	private void setTaskState(int i) {
		taskProgress = Math.max(Math.min(i, 100), 0);
		return; }
    
    /**
	 * This method is used to update the informal state of the acutal task by the specified improvement.
	 *  
	 * @param diff the progress of the acutal task.
	 * @return the previous percentage of the actual task.
	 */
	private int increaseTaskProgress(int diff) {
		taskProgress = Math.max(Math.min(taskProgress + diff, 100), 0);
		return taskProgress; }
	
	/**
	 * This method starts the Chinese Whispers clustering algorithm for the specified graph.
	 * 
	 * @param fgraph
	 * @param fout
	 * @param fclusters
	 * @deprecated
	 */
	@SuppressWarnings("unused")
	private void runCWAlgorithm(String fgraph, String fout, String fclusters, String fwlist) {
		printInfoAlienOutputBegin("CW output");
		ChineseWhispers cw = new ChineseWhispers();
		cw.isNumbered    = true; /* true !!!test!!! */
		cw.graphInMemory = false;
		cw.setCWGraph(fwlist, fgraph);
		cw.setCWParameters(0, "top", "", 0.0, "constant", 0.0, "continuous", 20, true);
        cw.run();
        cw.consoleChooserString_colors      = fout;
        cw.consoleChooserString_colors_read = fclusters;
        cw.writeFile(false, true, true);
        cw = null;
        printInfoAlienOutputEnd();
		return; }
	
	/**
	 * This method moves a file from <code>from</code> to <code>to</code>.
	 * @param from
	 * @param to
	 * @throws Exception
	 */
	@SuppressWarnings("unused")
	private void moveFile(String from, String to) throws Exception {
		File fromFile = new File(from);
		File toFile   = new File(to);
		
		if (toFile.exists() && !toFile.delete()) {
			throw new Exception("File " + fromFile.getPath() + " cannot be moved to " + toFile.getPath() +
					            " (" + toFile.getPath() + " already exists)."); }
		if (!fromFile.renameTo(toFile)) {
			throw new Exception("File " + fromFile.getPath() + " cannot be moved to " + toFile.getPath()); }
		return; }
	
	private void copyFile(String from, String to) throws Exception {
		try {
			BufferedInputStream  fin  = new BufferedInputStream(new FileInputStream(from));
			BufferedOutputStream fout = new BufferedOutputStream(new FileOutputStream(to));
			int c;
			while ((c = fin.read()) != -1) {
				fout.write(c); }
			fin.close();
			fout.close(); }
		catch (Exception e) {
			throw new Exception(e); }
		return;	}
	
	/**
	 * A description of the actual step (preprocessing, partitioning 1, ...
	 */
	private String stepName;
	
	/**
	 * A description of the actual task.
	 */
	private String taskName;
	
	/**
	 * The progress of the actual task (between 0 and 100).
	 */
	private int taskProgress = 0;
	
	/**
	 * Specifies case sensitivity of the <code>UnsuPosTagger</code>.
	 */
	private boolean ignoreCase = false;

	/**
	 * Controls preprocessing (options are false, true, break, force).
	 */
	private String doPreproc;
	
	/**
	 * Controls which tokenizer is used ("medusa" - Medusa, "normal"). 
	 */
	private String tokImpl;
	
	/**
	 * Controls partitioning 1 (options are false, true, break, force).
	 */
	private String doPart1;
	
	/**
	 * Controls partitioning 2 (options are false, true, break, force).
	 */
	private String doPart2;
	
	/**
	 * Controls whether partitioning 1 and 2 are joined after creation.
	 */
	private String doJoinParts;
	
	/**
	 * Controls whether lexicon is built.
	 */
	private String doBuildLex;
	
	/**
	 * Controls if tagging is performed (options are force and break).
	 */
	private String doTagg;
	
	/**
	 * Indicates if the <code>UnsuPosTagger</code> should stop as soon as possible.
	 */
	private boolean stopThread;
	
	/**
	 * The total running time of the unsupervised tagger.
	 */
	private Timer totalTime;
	
	private File dirBasic;
	private File dirGraph1;
	private File dirGraph2;
	private File dirGraph3;
	private File dirTmodel;
	private File dirLog;
	private File fcorpus;
	private File ftok;
	private File fwn;
	private File fwf;
	private File fcosgraph;
	private File fsim12;
	private File fsim21;
	private File fTmodel;
	private File fnodes;
	private File fedges;
	private File fcluster;
	private File fcr;
	
	private ArrayList<String> tmpFileList;
	private boolean deleteTempFiles;
	
	/**
	 * The new line sequence.
	 */
	private String NL = System.getProperty("line.separator");
	
	/**
	 * The output-<code>PrintStream</code>.
	 */
	private PrintStream sout;
	
	/**
	 * The error output <code>PrintStream</code>.
	 */
	private PrintStream eout;
	
	/**
	 * The settings read from the configuration file.
	 */
	private Properties settings = new Properties();
	
	/**
	 * A string.
	 */
	private String strInfoAlienOutput = "";
	
	/**
	 * Number of sentences (non-empty lines) in the corpus file (a value of -1 indicates an error).
	 */
	private int Nsmax;
	
	/**
	 * The number of feature words (see parameter <code>FEAT</code> in the configuration file).
	 */
	private int Nwfeat;
	
	/**
	 * The number of target words (see parameter <code>TARG</code> in the configuration file).
	 */
	private int Nwtarg;
	
	/**
	 * The number of words in the corpus file (i.e. the greatest word number).
	 */
	private int Nwmax = -1;
	
	/**
	 * The threshold value for cosinus of angle between context vectors (see <code>COS_THRESH</code> in
	 * the configuration file).
	 */
	private double Tcos;
	
	/**
	 * NB_THRESH
	 */
	private int Tnb;
	
	/**
	 * NB_MAX
	 */
	private int Nnbmax;
	
	/**
	 * (see parameter <code>BEHEAD</code> in the configuration file).
	 */
	private int Nbehead;
	
	/**
	 * The minimal overlap between clusters from partitioning 1 and 2, respectively (see parameter
	 * <code>CONF_OVERLAP</code> in the configuration file).
	 */
	private int Noverlap;
	
	/**
	 * The number of target words for the Chinese Whispers clustering algorithm (see parameter
	 * <code>CWTARG</code> in the configuration file).
	 */
	private int NCWtarg;
	
	private int Ntop;
	
	/**
	 * The number of singeltons to add from partitioning 1 (see parameter <code>SING_ADD</code> in the
	 * configuration file).
	 */
	private int Nsingadd;
	
	/**
	 * The matrix which consists of context vectors for all target words (row i is the context vector
	 * for the i-th target word (used for partitioning 1).
	 */
	private int[][]  V    = null;
	
	/**
	 * The Euclidian length of the context vectors (used for partitioning 1).
	 */
	private double[] vnorm = null;
}
