/**
 * This file was adapted for the UnsuPos system. Some annotations <code>@SuppressWarnings("unchecked")</code>
 * were added. Furthermore the <code>printDBResults()</code> method of the <code>RankedResultList</code>
 * class was modified to write only necessary information.
 * 
 * @author Andreas Klaus 
 */

package de.uni_leipzig.asv.unsupos.similarity;

import java.util.*;
import de.uni_leipzig.asv.coocc.*;


/**
 * Keeps track of possible words which could have a non-zero value of similarity.
 * Basically these are all direct collocations and their collocations.
 * Returns these in batches, of given size.
 * Uses the bitfield in order to recognize already calculated pairs to not
 * calculate (and print) them twice
 *
 * @author Stefan Bordag
 * @date   15.11.2003
 */
final public class Candidates
{
  protected Integer wordNr = null;

  /**
   * This bitfield contains information about which words were already done
   * earlier so that they don't have to be calculated again
   */
  protected BitField doneNumbers = null;

  protected BinFileMultCol cooccs = null;

  // if this is set to false, it will ignore the collocations of collocations
  protected boolean takeCollocOfColloc = true;

  /**
   * These are the cooccurrences of the initial word, they will be kept as long as
   * this word is actual.
   */
  protected List mainCooccs = null;

  /**
   * Once all main Cooccs are done, this will be filled with the cooccs of the
   * first word in mainCooccs and processed, then from the second and so on.
   */
  protected List furtherCooccs = null;

  protected int mainCounter = 0;
  protected int furtherCounter = 0;
  protected boolean mainDone = false;

  protected int maxItems = 0;
  protected int[] minCols = null;
  protected int[] maxCols = null;

  protected Integer nextCandidate = null;

  protected Set locallyDoneNumbers = null;
  
  private SimOptions simopt = null;

  public Candidates(SimOptions simopt, Integer wordNr, BitField doneWords, List cooccsOfInput, BinFileMultCol cooccs, Set locallyDoneNums)
  {
	this.simopt = simopt;
    this.wordNr = wordNr;
    this.doneNumbers = doneWords;
    this.mainCooccs = cooccsOfInput;
    this.cooccs = cooccs;
    this.takeCollocOfColloc = this.simopt.getTakeFurther();
    this.minCols = this.simopt.getMinCols();
    this.maxCols = this.simopt.getMaxCols();
    this.maxItems = this.simopt.getMaxCooccsCandidate();

    this.locallyDoneNumbers = locallyDoneNums;

    this.nextCandidate = getMyNextCandidate();
    while ( this.nextCandidate != null && ( this.nextCandidate.intValue() == this.wordNr.intValue()
           || this.doneNumbers.getPosition(this.nextCandidate.intValue())
           || this.locallyDoneNumbers.contains(this.nextCandidate)) )
    {
//      System.out.println("A locDoneNs: "+this.locallyDoneNumbers);
//      System.out.println("A locDoneNs.contains("+this.nextCandidate+")="+this.locallyDoneNumbers.contains(this.nextCandidate));
      this.nextCandidate = getMyNextCandidate();
    }
  }

  /**
   * Returns true as long as there are some elements left, false otherwise.
   * @return boolean
   */
  public boolean hasNextCandidate()
  {
    if ( nextCandidate == null )
    {
      return false;
    }
    return true;
  }

  @SuppressWarnings("unchecked")
  public Integer getNextCandidate()
  {
    Integer retVal = this.nextCandidate;
//    System.out.println("Going to return "+retVal+"");
    this.nextCandidate = getMyNextCandidate();
//    System.out.println("Is "+this.nextCandidate+" then a good next candidate?");
    this.locallyDoneNumbers.add(retVal);
    while ( this.nextCandidate != null && ( this.nextCandidate.intValue() == this.wordNr.intValue()
            || this.doneNumbers.getPosition(this.nextCandidate.intValue())
            || this.locallyDoneNumbers.contains(this.nextCandidate) ) )
    {
//      System.out.println("B locDoneNs: "+this.locallyDoneNumbers);
//      System.out.println("B locDoneNs.contains("+this.nextCandidate+")="+this.locallyDoneNumbers.contains(this.nextCandidate));
      this.nextCandidate = getMyNextCandidate();
    }
    return retVal;
  }

  /**
   * As long as the mains are not all used, give the next and increase count.
   * If the count reaches the number of mains, set it to 0 and begin returning
   * the further candidates.
   * From then on the mainCounter is used to see whose cooccurrences are
   * currently being processed. The furtherCounter is then used like the
   * mainCounter before.
   * @return Integer
   */
  public Integer getMyNextCandidate()
  {
    Integer retVal = null;
    if ( !this.mainDone )
    {
      if ( this.mainCounter >= this.mainCooccs.size() )
      {
        this.mainDone = true;
        this.mainCounter = 0;
        return getMyNextCandidate();
      }
      retVal = ((Integer[])this.mainCooccs.get(this.mainCounter))[0];
      this.mainCounter++;
      return retVal;
    }
    else
    {
      // all words are done, return null (why was this method then called?)
      if ( this.mainCounter > this.mainCooccs.size() )
      {
        //System.out.println("Warning, use hasNextCandidate() to not call this method in this state!");
        return null;
      }
      // current further batch done, fetch new one
      if ( this.furtherCooccs == null || this.furtherCounter >= this.furtherCooccs.size() )
      {
        if ( this.furtherCooccs != null )
        {
          if ( mainCounter >= mainCooccs.size() )
          {
            return null;
          }
        }
        //System.out.println("For candidate: "+this.wordNr+" got mainCooccs: "+this.mainCooccs);
        //System.out.println("size: "+this.mainCooccs.size());
        //System.out.println("size: "+this.mainCooccs.size());
        Integer curMainWord = ((Integer[])this.mainCooccs.get(this.mainCounter))[0];
        // here skip as many words as are already processed since their
        // cooccurrences have already been compared with the current word when
        // they were being processed
        this.furtherCooccs = this.cooccs.getData(curMainWord, SimOptions.cols, this.maxItems, this.minCols, this.maxCols);
        this.furtherCounter=0;
        this.mainCounter++;
        return getMyNextCandidate();
      }
      else
      {
        retVal = ((Integer[])this.furtherCooccs.get(this.furtherCounter))[0];
        this.furtherCounter++;
        return retVal;
      }
    }
  }

  public boolean getMainDone()
  {
    return this.mainDone;
  }

  private String printIntArray(Integer[] array)
  {
    String retVal = "(";
    for ( int i = 0 ; i < array.length ; i++ )
    {
      retVal = retVal + "" + array[i] + ", ";
    }
    return retVal+")";
  }

  public String toString()
  {
    String retString = "KollokCandidates: \n";
    retString = retString + " " + this.mainDone+"\n";
    retString = retString + " " + this.mainCounter+" / "+this.mainCooccs.size()+"\n";
    retString = retString + " " + this.furtherCounter+"\n";
    retString = retString + " mainCoocs: ";
    for ( Iterator it = this.mainCooccs.iterator() ; it.hasNext() ; )
    {
      Integer[] curCoocs = (Integer[])it.next();
      retString = retString + printIntArray(curCoocs);
    }
    retString = retString + "\n ";
    if ( this.furtherCooccs == null )
    {
      return retString;
    }
    retString = retString + " furtherCooccs: ";
    for ( Iterator it = this.furtherCooccs.iterator() ; it.hasNext() ; )
    {
      Integer[] curCoocs = (Integer[])it.next();
      retString = retString + printIntArray(curCoocs);
    }
    return retString;
  }

}
