/**
 * This file was adapted for the UnsuPos system. Some annotations <code>@SuppressWarnings("unchecked")</code>
 * were added. Furthermore the <code>printDBResults()</code> method of the <code>RankedResultList</code>
 * class was modified to write only necessary information.
 * 
 * @author Andreas Klaus 
 */

package de.uni_leipzig.asv.unsupos.similarity;

//import de.uni_leipzig.asv.wordserver.*;
import de.uni_leipzig.asv.coocc.*;

import java.util.*;

/**
 * In order to not be forced to compare every item with every other, a cue
 * source is needed which tells which items might be candidates to compare
 * with. In a graph, comparing by neighbours, candidates for example can only
 * be neighbours and their neighbours. By definition no other node of that
 * given graph will have a node in common with the initial node.
 *
 * Since there might be several different feature sources for an item (like
 * left neighbours, right neighbours, etc.) the algorithm is programmed to use
 * a merge method from Merger in order to create one vector representing all
 * these features.
 *
 * @author Stefan Bordag
 */
public class SimAlgorithm
{
  /**
   * This is the source of information on which candidates might be good.
   * from this, the first and second order neighbours will be fetched. (As
   * specified in the options, such that a direct candidate source might be
   * used)
   */
  protected BinFileMultCol cueSource = null;

  /**
   * From here the features will be taken, merged in the Merger class and the
   * used to compare items which came from the cueSource
   */
  protected BinFileMultCol[] featureSources = null;

  /**
   * This one is modified by the Similarity class and used by this class in
   * order to keep track which words to ignore because they have already been
   * computed.
   */
  protected BitField doneWords = null;

  /**
   * This stores the integers of the locally already done numbers in order
   * to avoid double similarity computation.
   * FIXME: Find out whether this actually improves or slows down the process
   */
  protected Set locallyDoneNumbers = null;

  protected int minAnzahl = 0;

  private SimOptions simopt = null;
  
  public SimAlgorithm(SimOptions simopt, BitField doneWords, BinFileMultCol cueSource, BinFileMultCol[] featureSources)
  {
	this.simopt    = simopt;
    this.minAnzahl = this.simopt.getMinAnzahl();
    this.doneWords = doneWords;
    this.cueSource = cueSource;
    this.featureSources = featureSources;
    this.locallyDoneNumbers = new HashSet(this.simopt.getMaxCooccsCandidate()*this.simopt.getMaxCooccsCandidate());

  }

  public SimAlgorithm(SimOptions simopt, BitField doneWords, BinFileMultCol cueSource, BinFileMultCol featureSource)
  {
	this.simopt    = simopt;
    this.minAnzahl = this.simopt.getMinAnzahl();
    this.doneWords = doneWords;
    this.cueSource = cueSource;
    this.featureSources = new BinFileMultCol[1];
    this.featureSources[0] = featureSource;
    this.locallyDoneNumbers = new HashSet(this.simopt.getMaxCooccsCandidate()*this.simopt.getMaxCooccsCandidate());
  }

  public SimAlgorithm(SimOptions simopt, BitField doneWords, BinFileMultCol cueAndFeatureSource)
  {
	this.simopt    = simopt;
    this.minAnzahl = this.simopt.getMinAnzahl();
    this.doneWords = doneWords;
    this.cueSource = cueAndFeatureSource;
    this.featureSources = new BinFileMultCol[1];
    this.featureSources[0] = cueAndFeatureSource;
    this.locallyDoneNumbers = new HashSet(this.simopt.getMaxCooccsCandidate()*this.simopt.getMaxCooccsCandidate());
  }


  @SuppressWarnings("unchecked")
  protected RankedResultList computeSimilarity(Integer wordNr)
  {
    // get all collocations and collocations of collocations in batches as candidates
    this.locallyDoneNumbers.clear();
    List da = this.cueSource.getData(wordNr, SimOptions.cols, simopt.getMaxCooccsCandidate(), simopt.getMinCols(), simopt.getMaxCols());

    if ( da == null || da.size() < 1 )
    {
      return new RankedResultList(simopt);
    }
    Candidates candidates = new Candidates(simopt, wordNr, this.doneWords, da, this.cueSource, this.locallyDoneNumbers);

    Map featuresOfInput = getFeatures(wordNr);

    RankedResultList resultList = new RankedResultList(simopt);

    HashSet localDoneWords = new HashSet();

    //System.out.println("BEGIN:\n----------\n"+candidates+"\n----------");

    double[] sim = null;
    Integer curCandidate = null;
    Map curCandidateFeatures = null;
    while ( candidates.hasNextCandidate() )
    {
      curCandidate = candidates.getNextCandidate();
      //System.out.println("Got new curCandidate: "+curCandidate);
      //System.out.println("proceding:\n----------\n"+candidates+"\n----------");
      if ( curCandidate == null )
      {
        break;
      }
/*      if ( this.doneWords.getPosition(curCandidate.intValue()) ||
           ( candidates.getMainDone() && featuresOfInput.containsKey(curCandidate) ) ||
           localDoneWords.contains(curCandidate) ) //don't calculate things twice
      {
        System.out.println("NOT Calc'ing between  "+wordNr+" and "+curCandidate);
        continue;
      }*/
      if ( simopt.debug ) { System.out.println("Calc'ing between  "+wordNr+" and "+curCandidate); }

      curCandidateFeatures = getFeatures(curCandidate);

      // but not to forget to remove the selfreferences...
      sim = SimilarityFormulae.calcSim(simopt, featuresOfInput,curCandidateFeatures);
      //System.out.println("Computed this sim:["+sim[SimOptions.BASELINE]+"] between "+wordNr+" and "+curCandidate+" ");
      if ( sim[SimOptions.BASELINE] > minAnzahl && sim[SimOptions.COS] > 0.001 )
      {
        resultList.putNextResult(wordNr, curCandidate, sim);
      }
      localDoneWords.add(curCandidate);
    }
    return resultList;
  }

  /**
   * Returns the merged features of a given word(number)
   * @param word Integer
   * @return Map
   */
  protected Map getFeatures(Integer wordNr)
  {
    Map[] m = new Map[this.featureSources.length];
    for ( int i = 0 ; i < this.featureSources.length ; i++ )
    {
      m[i] = this.featureSources[i].getDataAsMap(wordNr, SimOptions.cols, simopt.getMaxCooccsFeatures(), simopt.getMinCols(), simopt.getMaxCols());
    }

    return Merger.mergeToLastMap(m);
  }

  public static void printMap(int n,Integer wordNr, Map m)
  {
    System.out.println("Feature "+n+" of "+wordNr+" is: ");
    for ( Iterator it = m.keySet().iterator() ; it.hasNext() ; )
    {
      Integer curKey = (Integer)it.next();
      Integer[] curVals = (Integer[])m.get(curKey);
      System.out.print(curKey+"\t");
      for ( int i = 0 ; i < curVals.length ; i++ )
      {
        System.out.print(curVals[i]+"\t");
      }
      System.out.println();
    }
  }

}
