/**
 * This file was adapted for the UnsuPos system. Some annotations <code>@SuppressWarnings("unchecked")</code>
 * were added. Furthermore the <code>printDBResults()</code> method of the <code>RankedResultList</code>
 * class was modified to write only necessary information.
 * 
 * @author Andreas Klaus 
 */

package de.uni_leipzig.asv.unsupos.similarity;

import java.util.*;

/**
 * This class holds the formula(e) to compute similarity
 * @author Stefan Bordag
 */
public class SimilarityFormulae
{
  protected SimilarityFormulae() {  }

  /**
   * This method is a mess because it computes all measures at once in order to save
   * computational time. This makes sense because for most measures two or three
   * passes through all features are necessary and doing them separately would
   * mean as much unnecessary overhead as there are measures. Currently there
   * are 9 different measures, thus this save already an order of magnitude
   * computational time.
   * 0 = anzahl   Summe(signum(abs(Ai*Bi)))
   * 1 = anz_norm Summe(signum(abs(Ai*Bi)))/(max(Laenge(A),Laenge(B))
   * 2 = Dice_bin     2*anzahl/(Anzahl(A)+Anzahl(B))
   * 3 = Jaccard  anzahl/(Anzahl(A)+Anzahl(B)-anzahl)     auch manchmal Tanimoto Abstand genannt
   * 4 = cos      A*B/(|A|*|B|)
   * 5 = 1-norm   |(A1-B1)| + |(A2-B2)| + |(A3-B3)| + ...  City-block metrik oder manhattan distance
   * 6 = 2-norm   |A-B| = sqrt((A1-B1)^2 + (A2-B2)^2 + ... )
   * 7 = cos_bin  anzahl / Wurzel(Anzahl(A),Anzahl(B))
   * // 8 = Jennsen Shannon divergence = JS(A,B) = log(2) + 0.5 * sum ( h(sig(C,A)+sig(C,B)) - h(sig(C,A)) - h(sig(sig(C,B)) ) with h(x)=xlogx and sig(A,B)=p(B|A)
   * 8 = JS Divergence = JS(A,B) = 1/2 * ( D(q||avg(q,r)) + D(r||avg(q,r)) ) with avg(q,r)=(q(y)+r(y))/2
   * @param collocs Map
   * @param collocs2 Map
   * @return double[]
   */
  @SuppressWarnings("unchecked")
  public static double[] calcSim(SimOptions simopt, Map collocs, Map collocs2)
  {
    int factor = simopt.getFactor();
    int inputFactor = simopt.getInputDataFactor();
    Map newColls = new HashMap(collocs.size());

    //Map avgMapJS = new HashMap(); //
    Map probMapJSA = new HashMap(collocs.size()); //
    Map probMapJSB = new HashMap(collocs2.size()); //
    double sumJSA = 0.0;
    double sumJSB = 0.0;
    double sumJSAvg = 0.0;

    // this transforms all integer values into double values in collocs and normalizes with input Factor
    for ( Iterator it = collocs.keySet().iterator() ; it.hasNext() ; )
    {
      Integer curKey = (Integer) it.next();
      Integer[] curValues = (Integer[]) collocs.get(curKey);
      Double[] newValues = new Double[curValues.length];
      for ( int i = 0 ; i < curValues.length ; i++ )
      {
        newValues[i] = new Double(curValues[i].doubleValue()/inputFactor);
      }
      newColls.put(curKey, newValues);
      //avgMapJS.put(curKey,newValues[0]);
      sumJSA += newValues[0].doubleValue();
      probMapJSA.put(curKey,newValues[0]);
    }
    collocs = newColls;
    
    Map newColls2 = new HashMap(collocs2.size());
    // this transforms all integer values into double values in collocs2 and normalizes with input Factor
    for ( Iterator it = collocs2.keySet().iterator() ; it.hasNext() ; )
    {
      Integer curKey = (Integer) it.next();
      Integer[] curValues = (Integer[]) collocs2.get(curKey);
      Double[] newValues = new Double[curValues.length];
      for ( int i = 0 ; i < curValues.length ; i++ )
      {
        newValues[i] = new Double(curValues[i].doubleValue()/inputFactor);
      }
      newColls2.put(curKey, newValues);
      sumJSB += newValues[0].doubleValue();
      probMapJSB.put(curKey,newValues[0]);

      Double JSNewVal = null;
      if ( collocs.containsKey(curKey) )
      {
        JSNewVal = ((Double[])collocs.get(curKey))[0];
      }
      else
      {
        JSNewVal = new Double(0.0);
      }
      sumJSAvg += (JSNewVal.doubleValue()+newValues[0].doubleValue())/2.0;

    }
    collocs2 = newColls2;
    
    // Note: avgMapJS does not contain probabilities yet, as also avg does not

    // berechne eine Map mit den Raengen zu jedem Wort (sortieren, absteigend
    // immer dann naechsten Rang vergeben, wenn sich Zahl aendert)

    double[] sim = new double[SimOptions.MEASURES];

    Set completeList = new HashSet(collocs.size() + collocs2.size());

    // compute length of first vector
    double lengthA = 0.0;
    for (Iterator it = collocs.keySet().iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Double[] curValues = (Double[]) collocs.get(curKey);
      lengthA += Math.pow(curValues[0].doubleValue(), 2.0);
      completeList.add(curKey);
      // for BASELINE here:
      if ( collocs2.containsKey(curKey) )
      {
        sim[SimOptions.BASELINE] += 1.0;
      }
    }

    if ( sim[SimOptions.BASELINE] < 1.0 ) // make computations faster by quitting here
      // because anyway only dissimilarity will be computed
    {
      return sim;
    }
    lengthA = Math.sqrt(lengthA);

    // compute length of second vector
    double lengthB = 0.0;
    for (Iterator it = collocs2.keySet().iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Double[] curValues = (Double[]) collocs2.get(curKey);
      lengthB += Math.pow(curValues[0].doubleValue(), 2.0);
      completeList.add(curKey);
    }
    lengthB = Math.sqrt(lengthB);

    // do anz_norm here since we know already everything needed for that
    sim[SimOptions.BASELINE_NORM] = factor * (sim[SimOptions.BASELINE] / Math.min( (double) collocs.size(), (double) collocs2.size()));
    sim[SimOptions.COS_BIN] = (sim[SimOptions.BASELINE] / Math.sqrt( (double) collocs.size() * (double) collocs2.size())) * factor;


    // now actually do the computations
    // JS here too
    for (Iterator it = completeList.iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Double[] sigA = (Double[]) collocs.get(curKey);
      Double[] sigB = (Double[]) collocs2.get(curKey);
      double JSA = 0.0;
      double JSB = 0.0;
      //double JSAvg = 0.0;
      if (sigA != null && sigB != null)
      {
        sim[SimOptions.COS] = sim[SimOptions.COS] + sigA[0].doubleValue() * sigB[0].doubleValue();
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigA[0].doubleValue() - sigB[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigA[0].doubleValue() - sigB[0].doubleValue());
        //JSAvg = ((Double)avgMapJS.get(curKey)).doubleValue()/sumJSAvg;
        JSA = ((Double)probMapJSA.get(curKey)).doubleValue();
        JSB = ((Double)probMapJSA.get(curKey)).doubleValue();
      }
      else if (sigA != null)
      {
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigA[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigA[0].doubleValue());
        //JSAvg = ((Double)avgMapJS.get(curKey)).doubleValue()/sumJSAvg;
        JSA = ((Double)probMapJSA.get(curKey)).doubleValue();
      }
      else if (sigB != null)
      {
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigB[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigB[0].doubleValue());
        //JSAvg = ((Double)avgMapJS.get(curKey)).doubleValue()/sumJSAvg;
        JSB = ((Double)probMapJSB.get(curKey)).doubleValue();
      }
      sim[SimOptions.JENNSEN_SHANNON] += xlogxmy( JSA/sumJSA,((JSA+JSB)/2.0)/sumJSAvg ) + xlogxmy( JSB/sumJSB,((JSA+JSB)/2.0)/sumJSAvg );
    }
    sim[SimOptions.JENNSEN_SHANNON] = sim[SimOptions.JENNSEN_SHANNON] / 2.0;
    sim[SimOptions.JENNSEN_SHANNON] *= -1*factor;
    //sim[SimOptions.JENNSEN_SHANNON] = Math.abs(sim[SimOptions.JENNSEN_SHANNON]);
    sim[SimOptions.DICE] = ( (2.0 * sim[SimOptions.BASELINE]) / ( (double) collocs.size() + (double) collocs2.size())) *
      factor;
    if ( (double) collocs.size() + (double) collocs2.size() == sim[SimOptions.BASELINE])
    {
      sim[SimOptions.JACCARD] = 1.0 * factor;
    }
    else
    {
      sim[SimOptions.JACCARD] = (sim[SimOptions.BASELINE] /
                      ( (double) collocs.size() + (double) collocs2.size() -
                       sim[SimOptions.BASELINE])) * factor;
    }
    sim[SimOptions.COS] = sim[SimOptions.COS] / (lengthA * lengthB) * factor;
    sim[SimOptions.CITY] = Math.abs(sim[SimOptions.CITY]);
    sim[SimOptions.EUKLID] = Math.sqrt(sim[SimOptions.EUKLID])*100;

    return sim;
  }

  // the same as above, but using Integer instead of the Double transformation
  // in the beginning
  @SuppressWarnings("unchecked")
  public static double[] calcSimInteger(SimOptions simopt, Map collocs, Map collocs2)
  {
    int factor = simopt.getFactor();

    // berechne eine Map mit den Raengen zu jedem Wort (sortieren, absteigend
    // immer dann naechsten Rang vergeben, wenn sich Zahl aendert)

    double[] sim = new double[SimOptions.MEASURES];

    Set completeList = new HashSet(collocs.size() + collocs2.size());

    // compute length of first vector
    double lengthA = 0.0;
    for (Iterator it = collocs.keySet().iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Integer[] curValues = (Integer[]) collocs.get(curKey);
      // normalize with input data factor here

      lengthA += Math.pow(curValues[0].doubleValue(), 2.0);
      completeList.add(curKey);

      // for JS and BASELINE here:
      if ( collocs2.containsKey(curKey) )
      {
        //common+=1.0;
        sim[SimOptions.BASELINE] += 1.0;
      }
    }

    if ( sim[SimOptions.BASELINE] < 1.0 ) // make computations faster by quitting here
      // because anyway only dissimilarity will be computed
    {
      return sim;
    }
    lengthA = Math.sqrt(lengthA);

    double[][] JSArray = new double[2][];
    JSArray[0] = new double[(int)sim[SimOptions.BASELINE]];
    JSArray[1] = new double[(int)sim[SimOptions.BASELINE]];
    double JSsum1 = 0.0;
    double JSsum2 = 0.0;

    // compute length of second vector
    double lengthB = 0.0;
    int JSi = 0;
    for (Iterator it = collocs2.keySet().iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Integer[] curValues = (Integer[]) collocs2.get(curKey);
      lengthB += Math.pow(curValues[0].doubleValue(), 2.0);
      completeList.add(curKey);

      // JS
      if ( collocs.containsKey(curKey) )
      {
        Integer[] curValues2 = (Integer[]) collocs.get(curKey);
         JSArray[0][JSi] = curValues[0].doubleValue();
         JSArray[1][JSi] = curValues2[0].doubleValue();
         JSsum1+=JSArray[0][JSi];
         JSsum2+=JSArray[1][JSi];
         JSi++;
      }
    }
    lengthB = Math.sqrt(lengthB);

    // do anz_norm here since we know already everything needed for that
    sim[SimOptions.BASELINE_NORM] = factor * (sim[SimOptions.BASELINE] / Math.min( (double) collocs.size(), (double) collocs2.size()));
    sim[SimOptions.COS_BIN] = (sim[SimOptions.BASELINE] / Math.sqrt( (double) collocs.size() * (double) collocs2.size())) * factor;

    // JS final computations here
    for ( int j = 0 ; j < JSArray[0].length ; j++ )
    {
      JSArray[0][j] = JSArray[0][j]/JSsum1;
      JSArray[1][j] = JSArray[1][j]/JSsum2;
      sim[SimOptions.JENNSEN_SHANNON] += h(JSArray[0][j]+JSArray[1][j]) - h(JSArray[0][j]) - h(JSArray[1][j]);
    }
    sim[SimOptions.JENNSEN_SHANNON] = log(2.0)+0.5*sim[SimOptions.JENNSEN_SHANNON];
    double JSBeta = 2.0;
    sim[SimOptions.JENNSEN_SHANNON] = factor * Math.pow(10,-1*JSBeta*sim[SimOptions.JENNSEN_SHANNON]);
    // JS done here

    // now actually do the computations
    for (Iterator it = completeList.iterator(); it.hasNext(); )
    {
      Integer curKey = (Integer) it.next();
      Integer[] sigA = (Integer[]) collocs.get(curKey);
      Integer[] sigB = (Integer[]) collocs2.get(curKey);
      if (sigA != null && sigB != null)
      {
        sim[SimOptions.COS] = sim[SimOptions.COS] + sigA[0].doubleValue() * sigB[0].doubleValue();
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigA[0].doubleValue() - sigB[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigA[0].doubleValue() - sigB[0].doubleValue());
      }
      else if (sigA != null)
      {
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigA[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigA[0].doubleValue());
      }
      else if (sigB != null)
      {
        sim[SimOptions.EUKLID] = sim[SimOptions.EUKLID] + Math.pow(sigB[0].doubleValue(), 2.0);
        sim[SimOptions.CITY] = sim[SimOptions.CITY] + Math.abs(sigB[0].doubleValue());
      }
    }
    sim[SimOptions.DICE] = ( (2.0 * sim[SimOptions.BASELINE]) / ( (double) collocs.size() + (double) collocs2.size())) *
      factor;
    if ( (double) collocs.size() + (double) collocs2.size() == sim[SimOptions.BASELINE])
    {
      sim[SimOptions.JACCARD] = 1.0 * factor;
    }
    else
    {
      sim[SimOptions.JACCARD] = (sim[SimOptions.BASELINE] /
                      ( (double) collocs.size() + (double) collocs2.size() -
                       sim[SimOptions.BASELINE])) * factor;
    }
    sim[SimOptions.COS] = sim[SimOptions.COS] / (lengthA * lengthB) * factor;
    sim[SimOptions.CITY] = Math.abs(sim[SimOptions.CITY]);
    sim[SimOptions.EUKLID] = Math.sqrt(sim[SimOptions.EUKLID]);

    return sim;
  }

  /**
   * Vektoren mussen in Wkten transformiert werden.
   * Sie muessen Werte zwischen 0 und 1 enthalten, sowie sich insgesamt zu 1
   * aufsummieren.
   * Dazu: zuerst aufsummieren: 1+3+5+9 = 18
   * dann jeweils durch summe dividieren: 1/18 3/18 5/18 9/18
   * @return double
   */
/*  protected double computeKullbackLeibler(Map cooccs, Map coocc2,double anzahl)
  {
    double beta = 2.0;
    double numOfFeatures = cooccs.size()+coocc2.size()-anzahl;
    double[][] featureVec = getJSFeatures(cooccs,coocc2,numOfFeatures);
    double sim1 = 0.0;
    double sim2 = 0.0;
    for ( int i = 0 ; i < featureVec[0].length ; i++)
    {
      sim1 += featureVec[0][i]*log(div(featureVec[0][i],featureVec[1][i]));
      sim2 += featureVec[1][i]*log(div(featureVec[1][i],featureVec[0][i]));
    }
    sim1 = Math.pow(10, -1*beta*sim1);
    sim2 = Math.pow(10, -1*beta*sim2);
    return (sim1+sim2)/2;
  }*/

/*  protected static double computeJennsenShannon(Map cooccs, Map coocc2,double anzahl)
  {
    double beta = 2.0;
    double[][] featureVec = getJSFeatures(cooccs,coocc2,anzahl);
    double sim = 0.0;
    for ( int i = 0 ; i < featureVec[0].length ; i++)
    {
      System.out.println(" features["+i+"]("+featureVec[0][i]+","+featureVec[1][i]+")");
      sim += h(featureVec[0][i]+featureVec[1][i]) - h(featureVec[0][i]) - h(featureVec[1][i]);
    }
    sim = log(2.0)+0.5*sim;
    return Math.pow(10,-1*beta*sim);
  }*/

  protected static double xlogxmy(double x,double y)
  {
    if ( x > 0.0 && y > 0.0 )
    {
      return -1 * x * ( Math.log( x ) - Math.log ( y ) );
    }
    else if ( x > 0.0 )
    {
      return -1 * x * ( Math.log( x ) );
    }
    return 0.0;
  }


  protected static double h(double x)
  {
    if ( x > 0.0 )
    {
      return -1 * x * Math.log ( x );
    }
    return 0.0;
  }

  protected static double log(double x)
  {
    if ( x <= 0.0 )
    {
      return 0;
    }
    return Math.log(x);
  }

  protected static double div(double x, double y )
  {
    if ( y == 0 )
    {
      return 0;
    }
    return x/y;
  }

  /**
   * Only needs to get common ones, but probabilities must be less or equal to
   * one and sum up to 1
   * One pass to get them all and produce both sums
   * Another pass to divide by the sums
   * @param cooccs Map
   * @param cooccs2 Map
   * @param trueSize double
   * @return double[][]
   */
/*  protected static double[][] getJSFeatures(Map cooccs, Map cooccs2, double trueSize)
  {
    double[][] arr = new double[2][];
    arr[0] = new double[(int)trueSize];
    arr[1] = new double[(int)trueSize];
    int i = 0;
    double sum1 = 0.0;
    double sum2 = 0.0;
    for ( Iterator it = cooccs.keySet().iterator() ; it.hasNext() ;)
    {
      Object key = it.next();
      if ( cooccs2.containsKey(key) )
      {
        Integer[] curValues1 = (Integer[]) cooccs.get(key);
        Integer[] curValues2 = (Integer[]) cooccs2.get(key);
        arr[0][i] = curValues1[0].doubleValue();
        arr[1][i] = curValues2[0].doubleValue();
        sum1+=arr[0][i];
        sum2+=arr[1][i];
        i++;
      }
    }
    for ( int j = 0 ; j < arr[0].length ; j++ )
    {
      arr[0][j] = arr[0][j]/sum1;
      arr[1][j] = arr[1][j]/sum2;
    }
    return arr;
  }*/

  /**
   *
   * 0 = anzahl   Summe(signum(abs(Ai*Bi)))
   * 1 = anz_norm Summe(signum(abs(Ai*Bi)))/(max(Laenge(A),Laenge(B))
   * 2 = Dice     2*A*B/(|A|+|B|)
   * 3 = Jaccard  A*B/(|A|+|B|-A*B)     auch manchmal Tanimoto Abstand genannt
   * 4 = cos      A*B/(|A|*|B|)
   * 5 = 1-norm   |(A1-B1)| + |(A2-B2)| + |(A3-B3)| + ...  City-block metrik oder manhattan distance
   * 6 = 2-norm   |A-B| = sqrt((A1-B1)^2 + (A2-B2)^2 + ... )
   * @param collocs Map
   * @param collocs2 Map
   * @return double[]
   */
 /*  public double[] calcSimOld(Map collocs, Map collocs2)
   {
     double[] sim = new double[7];

     Set completeList = new HashSet(collocs.size() + collocs2.size());

     // compute length of first vector
     double lengthA = 0.0;
     for ( Iterator it = collocs.keySet().iterator() ; it.hasNext() ; )
     {
       Integer curKey = (Integer)it.next();
       Integer[] curValues = (Integer[])collocs.get(curKey);
       lengthA += Math.pow(curValues[0].doubleValue(),2.0);
       completeList.add(curKey);
     }
     lengthA = Math.sqrt(lengthA);

     // compute length of second vector
     double lengthB = 0.0;
     for ( Iterator it = collocs2.keySet().iterator() ; it.hasNext() ; )
     {
       Integer curKey = (Integer)it.next();
       Integer[] curValues = (Integer[])collocs2.get(curKey);
       lengthB += Math.pow(curValues[0].doubleValue(),2.0);
       if ( completeList.contains(curKey) ) // this is all what's needed for anzahl
       {
         sim[BASELINE] += 1.0;
       }
       completeList.add(curKey);
     }
     lengthB = Math.sqrt(lengthB);

     // do anz_norm here since we know already everything needed for that
     sim[BASELINE_NORM] = this.factor*(sim[BASELINE]/Math.min((double)collocs.size(),(double)collocs2.size()));

     // now actually do the computations
     for ( Iterator it = completeList.iterator() ; it.hasNext() ; )
     {
       Integer curKey = (Integer)it.next();
       Integer[] sigA = (Integer[])collocs.get(curKey);
       Integer[] sigB = (Integer[])collocs2.get(curKey);
       if ( sigA != null && sigB != null )
       {
         sim[COS] = sim[COS] + sigA[0].doubleValue() * sigB[0].doubleValue();
         sim[EUKLID] = sim[EUKLID] + Math.pow( sigA[0].doubleValue() - sigB[0].doubleValue(), 2.0 );
         sim[CITY] = sim[CITY] + Math.abs( sigA[0].doubleValue() - sigB[0].doubleValue());
       }
       else if ( sigA != null )
       {
         sim[EUKLID] = sim[EUKLID] + Math.pow( sigA[0].doubleValue(), 2.0 );
         sim[CITY] = sim[CITY] + Math.abs(sigA[0].doubleValue());
       }
       else if ( sigB != null )
       {
         sim[EUKLID] = sim[EUKLID] + Math.pow( sigB[0].doubleValue(), 2.0 );
         sim[CITY] = sim[CITY] + Math.abs(sigB[0].doubleValue());
       }
     }
     sim[DICE] = (2.0*sim[COS]) / (lengthA + lengthB);
     sim[JACCARD] = Math.abs((sim[COS]) / (lengthA + lengthB - sim[COS]))*this.factor;
     sim[COS] = sim[COS]/(lengthA*lengthB)*this.factor;
     sim[CITY] = Math.abs(sim[CITY]);
     sim[EUKLID] = Math.sqrt(sim[EUKLID]);

     return sim;
   }*/

  @SuppressWarnings("unchecked")
  public static void main(String[] args)
  {
    Map coocc1 = new HashMap();
    Integer[] in1 = {new Integer(1)};
    coocc1.put(new Integer(3),in1);
    Integer[] in12 = {new Integer(5)};
    coocc1.put(new Integer(5),in12);
    Integer[] in13 = {new Integer(4)};
    coocc1.put(new Integer(7),in13);
    Integer[] in14 = {new Integer(1)};
    coocc1.put(new Integer(9),in14);
    Integer[] in15 = {new Integer(40)};
    coocc1.put(new Integer(11),in15);
    Integer[] in16 = {new Integer(30)};
    coocc1.put(new Integer(13),in16);


    Map coocc2 = new HashMap();
    Integer[] in2 = {new Integer(1)};
    coocc2.put(new Integer(2),in2);
    Integer[] in22 = {new Integer(50)};
    coocc2.put(new Integer(5),in22);
    Integer[] in23 = {new Integer(4)};
    coocc2.put(new Integer(7),in23);
    Integer[] in24 = {new Integer(1)};
    coocc2.put(new Integer(8),in24);
    Integer[] in25 = {new Integer(40)};
    coocc2.put(new Integer(11),in25);
    Integer[] in26 = {new Integer(30)};
    coocc2.put(new Integer(13),in26);

    System.out.println("Computing JS:");
    //System.out.println("JS(A,B) = "+computeJennsenShannon(coocc1,coocc2,4.0));
  }


}
